import enum


class MyEnumMeta(enum.EnumMeta):
    """
    It will be used to check if the value exist in enum
    """
    def __contains__(cls, item):
        try:
            cls(item)
        except ValueError:
            return False
        else:
            return True


class Entry(enum.Enum, metaclass=MyEnumMeta):
    """
    Use like Entry.Buy,
    to check 0 in Entry
    """
    Buy = 1,
    Sell = 2
    Buy_Bet = 3
    Sell_Bet = 4
    HOLD = 5
    EXIT = 6

    def __str__(self):
        return self.name



from datetime import datetime, timedelta
from dateutil.tz.tz import gettz


def getting_market_session(sydney_tz='+11', tokyo_tz='+9', london_tz='+0', newyork_tz='-5'):
    session_str = ""
    if (datetime.now(tz=gettz(f"UTC"+sydney_tz)).replace(tzinfo=None).time().hour >= 7 and datetime.now(tz=gettz(f"UTC"+sydney_tz)).replace(tzinfo=None).time().hour < 16) and (datetime.now(tz=gettz(f"UTC"+sydney_tz)).replace(tzinfo=None).weekday() % 7 >= 0 and datetime.now(tz=gettz(f"UTC"+sydney_tz)).replace(tzinfo=None).weekday() % 7 <= 4):
        # ! sydney session open from 7:AM TO 4:00 PM LOCAL FROM MONDAY TO FRIDAY
        session_str += "Sydney Session"

    if (datetime.now(tz=gettz(f"UTC"+tokyo_tz)).replace(tzinfo=None).time().hour >= 9 and datetime.now(tz=gettz(f"UTC"+tokyo_tz)).replace(tzinfo=None).time().hour < 18) and (datetime.now(tz=gettz(f"UTC"+tokyo_tz)).replace(tzinfo=None).weekday() % 7 >= 0 and datetime.now(tz=gettz(f"UTC"+tokyo_tz)).replace(tzinfo=None).weekday() % 7 <= 4):
        # ! sydney session open from 7:AM TO 4:00 PM LOCAL FROM MONDAY TO FRIDAY
        if session_str != "":
            session_str += " & "

        session_str += "Tokyo Session"

    if (datetime.now(tz=gettz(f"UTC"+london_tz)).replace(tzinfo=None).time().hour >= 8 and datetime.now(tz=gettz(f"UTC"+london_tz)).replace(tzinfo=None).time().hour < 17) and (datetime.now(tz=gettz(f"UTC"+london_tz)).replace(tzinfo=None).weekday() % 7 >= 0 and datetime.now(tz=gettz(f"UTC"+london_tz)).replace(tzinfo=None).weekday() % 7 <= 4):
        # ! sydney session open from 7:AM TO 4:00 PM LOCAL FROM MONDAY TO FRIDAY
        if session_str != "":
            session_str += " & "

        session_str += "London Session (IMP)"

    if (datetime.now(tz=gettz(f"UTC"+newyork_tz)).replace(tzinfo=None).time().hour >= 8 and datetime.now(tz=gettz(f"UTC"+newyork_tz)).replace(tzinfo=None).time().hour < 17) and (datetime.now(tz=gettz(f"UTC"+newyork_tz)).replace(tzinfo=None).weekday() % 7 >= 0 and datetime.now(tz=gettz(f"UTC"+newyork_tz)).replace(tzinfo=None).weekday() % 7 <= 4):
        # ! sydney session open from 7:AM TO 4:00 PM LOCAL FROM MONDAY TO FRIDAY
        if session_str != "":
            session_str += " & "

        session_str += "New York Session (IMP)"

    if session_str == "":
        session_str = "Market is Closed Right Now"
    else:
        if " & " in session_str:
            session_str += " are Open Right Now"
        else:
            session_str += " is Open Right Now"

    return session_str



from MT5TrdHelper.helpers.entries import Entry
from pymt5adapter.types import OrderSendResult, TradePosition
import pytz
import pandas as pd
import time
from typing import Callable, Union
from datetime import datetime, timedelta
from MT5TrdHelper.exchange.symbol_base import Symbol
from MT5TrdHelper.exchange.mt5_exchange import MT5Exchange
# from helpers.database_helpers import get_hidden_be_db, get_hidden_sl_db, get_hidden_tp_db, get_hidden_trl_sl_db, insert_hidden_be_sl_db, insert_hidden_sl_db, insert_hidden_tp_db, insert_update_symbol_wave_info, insert_updt_hidden_trl_sl_db
import pymt5adapter as mt5
from pymt5adapter.const import ORDER_TIME, ORDER_TYPE, TIMEFRAME_M1, TRADE_ACTION, TRADE_RETCODE
from pymt5adapter.const import POSITION_TYPE

MARKET_CLOSED = 'Market closed'


class Trade:

    def __init__(self, exchange: MT5Exchange, symbol: str, magic_number=123456, slipaage=10, terminal='real_terminal_connection'):
        """
        initialize trade class
        exchange: an exchange base class object
        magic number = for identifying bot
        slippage: as deviation in the order functions
        """
        self.magic_number = magic_number
        self.slipaage = slipaage
        self.exchange = exchange
        self.TERMINAL = terminal
        self.symbol_name = symbol
        # ! Global variables
        self.FirstTick = True
        self.InitialCapital = 0
        self.profitAndLoss = 0
        self.output = False
        self.HiddenSLList = []
        self.HiddenTPList = []
        self.HiddenBEList = []
        self.HiddenTrailingList = []
        self.VolTrailingList = []
        self.HiddenVolTrailingList = []

    async def close_position(self, position: TradePosition) -> OrderSendResult:
        """
        Function to close an open position with TradePosition object
        """
        exc = self.exchange
        order_type = mt5.ORDER_TYPE.BUY if position.type == mt5.ORDER_TYPE.SELL else mt5.ORDER_TYPE.SELL
        position = exc.get_position_with_ticket_id(position.ticket)
        try:
            result = await exc.market_order_send(order_type=order_type, symbol=position.symbol,
                                                 volume=position.volume, position_ticket=position.ticket, slippage=self.slipaage, magic=self.magic_number, price=None)
            if result.retcode == mt5.TRADE_RETCODE.DONE:
                print("Successfully Closed Position: ", result.request.position)
            return result
        except Exception as ex:
            print(ex.args)

    async def close_position_volume(self, position: mt5.TradePosition, volume: float) -> mt5.OrderSendResult:
        """
        Function to close an open position with TradePosition object
        """
        exc = self.exchange
        order_type = mt5.ORDER_TYPE.BUY if position.type == mt5.ORDER_TYPE.SELL else mt5.ORDER_TYPE.SELL
        try:
            result = await exc.market_order_send(order_type=order_type, symbol=position.symbol,
                                                 volume=volume, position_ticket=position.ticket, slippage=self.slipaage, magic=self.magic_number, price=None)
            if result.retcode == mt5.TRADE_RETCODE.DONE:
                print("Successfully Closed Position: ", result.request.position)
            return result
        except Exception as ex:
            print(ex.args)

    async def close_all_position(self) -> None:
        """
        Function to close an open position with TradePosition object
        """
        exc = self.exchange
        magic = self.magic_number
        positions = exc.get_position_with_magic(magic)
        for pos in positions:
            order_type = ORDER_TYPE.BUY if pos.type == ORDER_TYPE.SELL else ORDER_TYPE.SELL
            try:
                result = await exc.market_order_send(order_type=order_type, symbol=pos.symbol,
                                                     volume=pos.volume, position=pos.ticket, slippage=self.slipaage, magic=self.magic_number)

            except Exception as ex:
                print(ex.args)

    async def close_all_buy_position(self) -> None:
        """
        Function to close an open position with TradePosition object
        """
        exc = self.exchange
        magic = self.magic_number
        positions = exc.get_position_with_magic(magic)
        buy_pos = filter(lambda p: p.type == POSITION_TYPE.BUY, positions)
        for pos in buy_pos:
            order_type = ORDER_TYPE.SELL
            try:
                result = await exc.market_order_send(order_type=order_type, symbol=pos.symbol,
                                                     volume=pos.volume, position=pos.ticket, slippage=self.slipaage, magic=self.magic_number)

            except Exception as ex:
                print(ex.args)

    async def close_all_sell_position(self) -> None:
        """
        Function to close an open position with TradePosition object
        """
        exc = self.exchange
        magic = self.magic_number
        positions = exc.get_position_with_magic(magic)
        buy_pos = filter(lambda p: p.type == POSITION_TYPE.SELL, positions)
        for pos in buy_pos:
            order_type = ORDER_TYPE.BUY
            try:
                result = await exc.market_order_send(order_type=order_type, symbol=pos.symbol,
                                                     volume=pos.volume, position=pos.ticket, slippage=self.slipaage, magic=self.magic_number)

            except Exception as ex:
                print(ex.args)

    async def close_all_pending_position(self, parameter_list):
        """
        Closes all pending position
        """
        exc = self.exchange
        for pos in exc .get_all_pending_position():
            await self.close_position(pos)

        assert len(exc .get_all_pending_position(
        )) == 0, 'Some pending Position may not have closed, take manual action'

    # ! place order via open market order (BUY SELL FUNCTION BELOW NOT NEEDED ANY MORE)
    # def buy(self, symbol: str, volume: float, sl_ticks: Union[int, float, None], tp_ticks: Union[int, float, None],) -> mt5.TradePosition:
    #     # TODO Implement P in sl & tp
    #     exc = self.exchange
    #     slippage = self.slipaage
    #     magic = self.magic_number

    #     tick = exc.get_symbol_info_tick(symbol)
    #     tick_size = exc.get_symbol_info(symbol).trade_tick_size
    #     sl = tick.ask - sl_ticks * tick_size if sl_ticks else None
    #     tp = tick.ask + tp_ticks * tick_size if tp_ticks else None
    #     result = exc.market_order_send(symbol=symbol, volume=volume, stop_loss=sl, take_profit=tp,
    #                                    order_type=mt5.ORDER_TYPE.BUY, slippage=slippage, magic=magic)
    #     assert result.comment != MARKET_CLOSED, 'Market is closed trade at a later time'
    #     pos = exc.get_position_from_order_send_result(result)
    #     return pos

    # def sell(self, symbol: str, volume: float, sl_ticks: int = None, tp_ticks: int = None) -> mt5.TradePosition:
    #     # TODO Implement P in sl & tp
    #     exc = self.exchange
    #     slippage = self.slipaage
    #     magic = self.magic_number

    #     tick = exc.get_symbol_info_tick(symbol)
    #     tick_size = exc.get_symbol_info(symbol).trade_tick_size

    #     sl = tick.bid + sl_ticks * tick_size if sl_ticks else None
    #     tp = tick.ask - tp_ticks * tick_size if tp_ticks else None
    #     result = exc.market_order_send(symbol=symbol, volume=volume, stop_loss=sl, take_profit=tp,
    #                                    order_type=mt5.ORDER_TYPE.SELL, slippage=slippage, magic=magic)

    #     pos = exc.get_position_from_order_send_result(result)
    #     assert pos.comment != MARKET_CLOSED, 'Market is closed trade at a later time'

    #     return pos

    def is_vol_limit_breached(self, symbol: str, VolLimitActivated: bool, VolMulti: float, ATR_val: float, debug: bool, entry_variable: Union[int, Entry]):
        """
        This function determines if our maximum volatility threshold is breached
        #  2 steps to this function:
        # --------
        #  1) It checks the price movement between current time and the closing price of the last completed 1min bar(shift 1 of 1min timeframe).
        #  2) Return True if this price movement > VolLimitMulti * VolATR
        """
        exc = self.exchange
        output = False
        one_min_sym = Symbol(symbol, TIMEFRAME_M1, terminal=self.TERMINAL)

        close = one_min_sym.close

        if(VolLimitActivated == False):
            return(output)

        priceMovement = abs(one_min_sym.bid-close)

        if(priceMovement > VolMulti*ATR_val):
            output = True
            if debug:
                if entry_variable in Entry:
                    print("Max volatility limit reached, Order won't be executed now")

        return(output)

    def is_max_pos_reached(self, max_positions_total: int, max_position_per_symbol: int, symbol_name: str, debug: bool, entry_variable: Union[int, Entry]):
        """
        This function checks the number of positions we are holding against the maximum allowed
        """
        result = False
        exc = self.exchange
        magic = self.magic_number
        positions = exc.get_position_with_magic(magic)
        pos = list(filter(lambda p: p.type == POSITION_TYPE.BUY or p.type ==
                          POSITION_TYPE.SELL, positions))
        symbol_pos = list(filter(lambda p: p.symbol == symbol_name, pos))

        if(len(pos) >= max_positions_total):
            result = True
            if(debug):
                if entry_variable in Entry:
                    print(
                        f"Max Orders Reached for this portoflio, no new order will be executed")

        elif len(symbol_pos) >= max_position_per_symbol:
            result = True
            if(debug):
                if entry_variable in Entry:
                    print(
                        f"Max Symbol Order Reached for symbol {symbol_name}, no new order will be executed")

        return(result)

    def modify_sl(self, ticket_id: int, SL: Union[int, float] = None) -> TradePosition:
        """
        modify sl for an already executed order
        """
        exc = self.exchange

        position = exc.get_position_with_ticket_id(ticket_id)
        symbol = Symbol(position.symbol, TIMEFRAME_M1, terminal=self.TERMINAL)

        if position.type == ORDER_TYPE.SELL:
            if isinstance(SL, int):
                stoploss = symbol.normalize_price(max(position.price_open, symbol.ask)+(SL *
                                                                                        symbol.p_value*symbol.point))
                stoploss = self.adjust_above_stop_level(
                    symbol, stoploss)
        if position.type == ORDER_TYPE.BUY:
            if isinstance(SL, int):
                SL = symbol.normalize_price(min(position.price_open, symbol.bid)-(SL *
                                                                                  symbol.p_value*symbol.point))
                SL = self.adjust_below_stop_level(symbol, SL)

        result = exc.modify_sl(stop_loss=SL, position=position)
        # ! checking if the retcode is done
        print(
            "Modify SL Result, ", result.comment)
        pos = exc.get_position_from_order_send_result(result)
        return pos

    def modify_tp(self, ticket_id: int, TP: Union[int, float] = None) -> TradePosition:
        """
        modify tp for an already executed order
        """
        exc = self.exchange

        position = exc.get_position_with_ticket_id(ticket_id)
        symbol = Symbol(position.symbol, TIMEFRAME_M1, terminal=self.TERMINAL)
        if position.type == mt5.ORDER_TYPE.SELL:
            if isinstance(TP, int):
                TP = symbol.bid - (TP * symbol.trade_tick_size)
            TP = self.adjust_below_stop_level(symbol, TP)
        if position.type == mt5.ORDER_TYPE.BUY:
            if isinstance(TP, int):
                TP = symbol.ask + (TP * symbol.trade_tick_size)
            TP = self.adjust_above_stop_level(symbol, TP)

        result = exc.modify_tp(take_profit=TP, position=position)
        # ! checking if the retcode is done
        print(
            "Modify TP Result, ", result.comment)
        pos = exc.get_position_from_order_send_result(result)
        return pos

    def adjust_below_stop_level(self, symbol: Symbol, stop_price: float, points=0):
        """
        adjust price below stop level
        """
        current_price = symbol.bid
        point = symbol.point
        # ! minimum stop loss required by the broker, don't have to apply p_value here
        stop_level = symbol.trade_stops_level * point
        calc_stop_price = current_price - stop_level
        add_points = points * point
        if stop_price < calc_stop_price - add_points:
            return stop_price
        else:
            new_price = calc_stop_price - add_points
            # print('Price adjusted below stop level to ', new_price)
            return new_price

    def adjust_above_stop_level(self, symbol: Symbol, stop_price, points=0):
        """
        adjust price above stop level
        """
        current_price = symbol.bid
        point = symbol.point
        # ! minimum stop loss required by the broker, don't have to apply p_value here
        stop_level = symbol.trade_stops_level * point
        calc_stop_price = current_price + stop_level
        add_points = points * point
        if stop_price > calc_stop_price + add_points:
            return stop_price
        else:
            new_price = calc_stop_price + add_points
            # print('Price adjusted above stop level to ', new_price)
            return new_price

    async def set_break_even_stop(self, ticket_id, min_profit_pip=0, lock_profit=0, break_even_close_half: bool = False):
        """
        Set break even stop for an order
        """
        exc = self.exchange

        assert min_profit_pip != 0, "minimum profit should be more than 0"

        result = exc.get_position_with_ticket_id(ticket_id)
        set_break_even_stop = False
        break_even_stop = 0

        # ! order and symbol information
        order_type = result.type
        order_symbol = result.symbol
        symbol = Symbol(order_symbol, TIMEFRAME_M1, terminal=self.TERMINAL)
        order_stop_loss = result.sl
        order_take_profit = result.tp
        order_open_price = result.price_open

        # ! only taking the p value

        # ! convert pips to prices
        min_profit = min_profit_pip * symbol.point * symbol.p_value

        lock_profit_val = 0
        if lock_profit > 0:
            lock_profit_val = lock_profit * symbol.point * symbol.p_value

        if order_type == ORDER_TYPE.BUY:
            break_even_stop = order_open_price + lock_profit_val
            break_even_stop = symbol.normalize_price(break_even_stop)
            # current_profit = exc.calc_order_profit(
            #     order_type, symbol.name, result.volume, order_open_price, symbol.bid)
            current_profit_pip = symbol.bid - order_open_price
            current_profit_pip = symbol.normalize_price(current_profit_pip)

            order_stop_loss = symbol.normalize_price(order_stop_loss)

            if break_even_stop > order_stop_loss and current_profit_pip >= min_profit:
                set_break_even_stop = True

        elif order_type == ORDER_TYPE.SELL:
            break_even_stop = order_open_price - lock_profit_val
            break_even_stop = symbol.normalize_price(break_even_stop)
            # current_profit = exc.calc_order_profit(
            #     order_type, symbol.name, result.volume, order_open_price, symbol.ask)
            current_profit_pip = order_open_price - symbol.ask
            current_profit_pip = symbol.normalize_price(current_profit_pip)

            order_stop_loss = symbol.normalize_price(order_stop_loss)
            if (break_even_stop < order_stop_loss or order_stop_loss == 0) and current_profit_pip >= min_profit:
                set_break_even_stop = True
        # ! set trailing step
        if set_break_even_stop:
            assert break_even_stop != 0, 'For some reason trail stop price haven\'t changed'
            result = exc.modify_sl(stop_loss=break_even_stop, position=result)

            # ! checking if the retcode is done
            assert result.retcode == mt5.TRADE_RETCODE.DONE
            pos = exc.get_position_from_order_send_result(result)
            # ! closing half of the trade on breakeven
            if break_even_close_half:
                half_vol = pos.volume * .5
                corrected_volume = symbol.verify_volume(
                    half_vol)
                if corrected_volume != pos.volume:
                    # ! we are not in the lowest volume closing half volume is possible
                    await self.close_position_volume(pos, corrected_volume)
                    self.CLOSE_HALF_PORTION = True
                else:
                    print(
                        f"We are in a lowest possible volume closing half volume is not possible for {symbol.name}")

    def set_break_even_stop_all(self, symbol: Symbol, min_profit, lock_profit=0):
        exc = self.exchange
        # magic = self.magic_number

        positions = exc.get_position_with_magic(self.magic_number)
        symbol_pos = list(filter(lambda p: p.symbol == symbol.name, positions))
        for position in symbol_pos:
            order_type = position.type
            if(order_type == ORDER_TYPE.BUY or order_type == ORDER_TYPE.SELL):
                self.set_break_even_stop(
                    position.ticket, min_profit, lock_profit)

    def is_loss_limit_breached(self, loss_limit_activated: bool, loss_limit_percentage: float, debug: bool, entry_signal_trigger: Union[int, Entry]):
        """
        This function determines if our maximum loss threshold is breached
        """
        initial_capital = self.exchange.get_current_capital(use_balance=True)
        if(loss_limit_activated == False):
            return(self.output)

        if(self.FirstTick == True):
            self.InitialCapital = initial_capital
            self.FirstTick = False

        profitAndLoss = (initial_capital/self.InitialCapital)-1

        if(profitAndLoss < -loss_limit_percentage/100):
            self.output = True
            profitAndLossPrint = round(profitAndLoss, 4)*100
            if(debug):
                if(entry_signal_trigger in Entry):
                    print("Entry trade triggered but not executed. Loss threshold breached. Current Loss: " +
                          str(profitAndLossPrint) + "%")

        return(self.output)

    def set_trailing_stop_all_price(self, symbol: Symbol, price, min_profit=0, step=10):
        """
        set trailing stop for all open order with price
        """
        exc = self.exchange
        magic = self.magic_number

        total_positions = exc.get_position_with_magic(self.magic_number)
        symbol_pos = list(filter(lambda p: p.symbol ==
                                 symbol.name, total_positions))
        for position in symbol_pos:
            order_type = position.type
            magic_number = position.magic
            if(magic_number == magic and (order_type == ORDER_TYPE.BUY or order_type == ORDER_TYPE.SELL)):
                self.set_trailing_stop_by_price(
                    position.ticket, price, min_profit, step)

    def set_vol_trailing_stop_all_price(self, symbol: Symbol, vol_value: float, min_profit: int = 0, step=10):
        """
        set trailing stop for all open order with price
        """
        exc = self.exchange
        magic = self.magic_number

        total_positions = exc.get_position_with_magic(self.magic_number)
        symbol_pos = list(filter(lambda p: p.symbol ==
                                 symbol.name, total_positions))
        for position in symbol_pos:
            order_type = position.type
            magic_number = position.magic
            if(magic_number == magic and (order_type == ORDER_TYPE.BUY or order_type == ORDER_TYPE.SELL)):
                if order_type == ORDER_TYPE.BUY:
                    price = symbol.bid - vol_value
                    self.set_trailing_stop_by_price(
                        position.ticket, price, min_profit, step)
                elif order_type == ORDER_TYPE.SELL:
                    price = symbol.ask + vol_value
                    self.set_trailing_stop_by_price(
                        position.ticket, price, min_profit, step)

    def set_trailing_stop_by_price(self, ticket_id, trail_price, min_profit, step=10):
        """
        Trailing stop price
        """
        exc = self.exchange
        magic = self.magic_number

        assert trail_price != 0 or trail_price != None, 'You must define trail price'
        set_trailing_stop = False

        result = exc.get_position_with_ticket_id(ticket_id)
        order_type = result.type
        order_symbol = result.symbol
        symbol = Symbol(order_symbol, TIMEFRAME_M1, terminal=self.TERMINAL)
        order_stop_loss = result.sl
        if order_stop_loss == 0:
            hidden_sl_list = self.HiddenSLList
            x = 0
            while x < len(hidden_sl_list):
                #  Looping through all order number in list
                orderTicketNumber = hidden_sl_list[x][0]
                orderSL = hidden_sl_list[x][1]
                if orderTicketNumber == ticket_id:
                    print(
                        'YOU SHOULDN\'T USE HIDDEN SL WITH TRAIL STOP POINT, EITHER USE BOTH HIDDEN OR BOTH REVEALED')
                    if order_type == ORDER_TYPE.BUY:
                        order_stop_loss = result.price_open - \
                            (orderSL * symbol.p_value * symbol.point)
                        break
                    elif order_type == ORDER_TYPE.SELL:
                        order_stop_loss = result.price_open + \
                            (orderSL * symbol.p_value * symbol.point)
                        break
                x = +1

        order_take_profit = result.tp
        order_open_price = result.price_open

        min_profit = min_profit * symbol.point*symbol.p_value
        step_value = step * symbol.point * symbol.p_value
        # if step < 10:
        #     step_value = 10 * symbol.point * symbol.p_value
        if order_type == ORDER_TYPE.BUY:
            trail_price = self.adjust_below_stop_level(symbol,
                                                       trail_price)

            # current_profit = exc.calc_order_profit(
            #     order_type, symbol.name, result.volume, order_open_price, symbol.bid)
            current_profit_pip = symbol.bid - order_open_price
            current_profit_pip = symbol.normalize_price(current_profit_pip)
            if trail_price > (order_stop_loss+step_value) and current_profit_pip >= min_profit:
                set_trailing_stop = True

        elif order_type == ORDER_TYPE.SELL:
            trail_price = self.adjust_above_stop_level(symbol,
                                                       trail_price)
            # current_profit = exc.calc_order_profit(
            #     order_type, symbol.name, result.volume, order_open_price, symbol.ask)
            current_profit_pip = order_open_price - symbol.ask
            current_profit_pip = symbol.normalize_price(current_profit_pip)
            if trail_price < (order_stop_loss-step_value or order_stop_loss == 0) and current_profit_pip >= min_profit:
                set_trailing_stop = True

        # ! set trailing step
        if set_trailing_stop:
            assert trail_price != 0, 'For some reason trail stop price haven\'t changed'
            pos = exc.modify_sl(stop_loss=trail_price, position=result)

            # ! checking if the retcode is done
            assert pos.retcode == mt5.TRADE_RETCODE.DONE, 'Trail Stop Setting In MT5 Caused some problem.'
            print(
                f"Trail Stop Updated for {pos.request.symbol}, to {trail_price}")

    def set_trailing_stop_all_point(self, symbol: Symbol, trail_points: int, min_profit: int, step=10):
        """
        set trailing stop for all open order with points
        """
        exc = self.exchange
        magic = self.magic_number
        total_positions = exc.get_position_with_magic(self.magic_number)
        symbol_pos = list(filter(lambda p: p.symbol ==
                                 symbol.name, total_positions))

        for position in symbol_pos:
            order_type = position.type
            magic_number = position.magic
            if(order_type == ORDER_TYPE.BUY or order_type == ORDER_TYPE.SELL):
                self.set_trailing_stop_by_point(
                    position.ticket, trail_points, min_profit, step)

    def set_trailing_stop_by_point(self, ticket_id, trail_points, min_profit, step=10):
        """
        set trailing stop with points
        """
        exc = self.exchange
        magic = self.magic_number

        assert trail_points != 0, 'vai set jokhon korbi to 0 disos kela'
        # ! initially kept set_trailing_stop to be false, and trail stop price to 0
        set_trailing_stop = False
        trail_stop_price = 0
        result = exc.get_position_with_ticket_id(ticket_id)
        order_type = result.type
        order_symbol = result.symbol
        symbol = Symbol(order_symbol, TIMEFRAME_M1, terminal=self.TERMINAL)
        order_stop_loss = result.sl
        if order_stop_loss == 0:
            hidden_sl_list = self.HiddenSLList
            x = 0
            while x < len(hidden_sl_list):
                #  Looping through all order number in list
                print(
                    'YOU SHOULDN\'T USE HIDDEN SL WITH TRAIL STOP POINT, EITHER USE BOTH HIDDEN OR BOTH REVEALED')
                doesOrderExist = False
                orderTicketNumber = hidden_sl_list[x][0]
                orderSL = hidden_sl_list[x][1]
                if orderTicketNumber == ticket_id:
                    if order_type == ORDER_TYPE.BUY:
                        order_stop_loss = result.price_open - \
                            (orderSL * symbol.p_value * symbol.point)
                        break
                    elif order_type == ORDER_TYPE.SELL:
                        order_stop_loss = result.price_open + \
                            (orderSL * symbol.p_value * symbol.point)
                        break
                x += 1
        order_take_profit = result.tp
        order_open_price = result.price_open

        # ! convert inputs to prices
        trail_points = trail_points * symbol.point * symbol.p_value
        min_profit = min_profit * symbol.point * symbol.p_value

        step_value = step * symbol.point * symbol.p_value

        # if step < 10:
        #     step_value = 10 * symbol.point * symbol.p_value

        if order_type == ORDER_TYPE.BUY:
            trail_stop_price = symbol.bid - trail_points
            trail_stop_price = symbol.normalize_price(trail_stop_price)

            # current_profit_pip = exc.calc_order_profit(
            #     order_type, symbol.name, result.volume, order_open_price, symbol.bid)
            current_profit_pip = symbol.bid - order_open_price
            current_profit_pip = symbol.normalize_price(current_profit_pip)
            if trail_stop_price > order_stop_loss+step_value and current_profit_pip >= min_profit:
                set_trailing_stop = True

        elif order_type == ORDER_TYPE.SELL:
            trail_stop_price = symbol.ask + trail_points
            trail_stop_price = symbol.normalize_price(trail_stop_price)
            # current_profit_pip = exc.calc_order_profit(
            #     order_type, symbol.name, result.volume, order_open_price, symbol.ask)
            current_profit_pip = order_open_price - symbol.ask
            current_profit_pip = symbol.normalize_price(current_profit_pip)
            if (trail_stop_price < order_stop_loss-step_value or order_stop_loss == 0) and current_profit_pip >= min_profit:
                set_trailing_stop = True

        if set_trailing_stop:
            assert trail_stop_price != 0, 'For some reason trail stop price haven\'t changed'
            pos = exc.modify_sl(stop_loss=trail_stop_price, position=result)

            # ! checking if the retcode is done
            assert pos.retcode == mt5.TRADE_RETCODE.DONE, 'Trail Stop Setting In MT5 Caused some problem.'
            print(
                f"Trail Stop Updated for {pos.request.symbol}, to {trail_stop_price}")

    # +------------------------------------------------------------------+
    # * (START) Set Hidden Stop Loss
    # +------------------------------------------------------------------+

    def set_stop_loss_hidden(self, debug: bool, SL: float, OrderNum: int, insert_hidden_sl_db: Callable):
        """
            insert_hidden_sl_db # this function should called db to insert the df
            This function calculates hidden stop loss amount and tags it to the appropriate order using an array
        """
        insert_hidden_sl_db(OrderNum, SL)
        if (debug):
            print(
                f"EA Debug: Order {str(OrderNum)} assigned with a hidden SL at {str(SL)}")

    # +------------------------------------------------------------------+
    #  * (END) Set Hidden Stop Loss
    #  +------------------------------------------------------------------+

    # +------------------------------------------------------------------+
    #  * (START) Set Hidden Take Profit
    # +------------------------------------------------------------------+
    def set_take_profit_hidden(self, debug: bool, TP: float, OrderNum: int, insert_hidden_tp_db: Callable):
        """
        insert_hidden_tp_db # this function should called db to insert the df
        This function calculates hidden take profit amount and tags it to the appropriate order using an array
        """
        insert_hidden_tp_db(OrderNum, TP)
        if (debug):
            print(
                f"EA Debug: Order {str(OrderNum)} assigned with a hidden tp at {str(TP)}")
    # +------------------------------------------------------------------+
    #  * (END) End of Set Hidden Take Profit
    # +------------------------------------------------------------------+

    # +------------------------------------------------------------------+
    #  * (START) Trigger Hidden Stop Loss
    # +------------------------------------------------------------------+
    async def trigger_stop_loss_hidden(self, debug: bool, symbol: Symbol, ticket: int, hidden_sl_df: pd.DataFrame):
        """
            hidden_sl_df = get_hidden_sl_db() # this function should called db to collect the df
            # This funciton is for Live Trading Only
            # This function does two 2 things:
            >>> 1) Clears appropriate elements of your HiddenSLList if positions has been closed
            >>> 2) Closes positions based on its hidden stop loss levels

        """
        exc = self.exchange
        pos = exc.get_position_with_ticket_id(ticket)
        # length = len(hidden_sl_list)
        hidden_sl_df = hidden_sl_df[hidden_sl_df['order_id'] == ticket]
        if pos and len(hidden_sl_df.index) != 0:
            hidden_sl = hidden_sl_df['stop_loss'].iloc[0]
            if pos.type == POSITION_TYPE.BUY and hidden_sl >= symbol.bid:
                close = True
            elif pos.type == POSITION_TYPE.SELL and hidden_sl <= symbol.ask:
                close = True
            else:
                close = False

            if close:
                if(debug):
                    print("EA Debug: Stop Loss hidden hit Trying to close position " +
                          str(ticket)+" ...")
                # HandleTradingEnvironment(debug, Retry_Interval);
                closing = await self.close_position(pos)
                # closing = OrderClose(
                #     OrderTicket(), OrderLots(), Bid, Slip*K, Blue);
                if(debug and not closing):
                    print(
                        "EA Debug: Unexpected Error has happened. While closing the order")
                elif(debug and closing):
                    print("EA Debug: Position successfully closed.")
    #  +------------------------------------------------------------------+
    #  * End of Trigger Hidden Stop Loss
    #  +------------------------------------------------------------------+

    # +------------------------------------------------------------------+
    #  * Trigger Hidden Take Profit
    # +------------------------------------------------------------------+

    async def trigger_take_profit_hidden(self, debug: bool, symbol: Symbol, ticket: int, hidden_tp_df: pd.DataFrame):
        """
        hidden_tp_df = get_hidden_tp_db() # this function should called db to collect the df
        # This function does two 2 things:
        >>> 1) Clears appropriate elements of your HiddenTPList if positions has been closed
        >>> 2) Closes positions based on its hidden take profit levels
        """
        exc = self.exchange
        pos = exc.get_position_with_ticket_id(ticket)

        hidden_tp_df = hidden_tp_df[hidden_tp_df['order_id'] == ticket]
        if pos and len(hidden_tp_df.index) != 0:
            take_profit = hidden_tp_df['take_prft'].iloc[0]
            if pos.type == POSITION_TYPE.BUY and take_profit <= symbol.bid:
                close = True
            elif pos.type == POSITION_TYPE.SELL and take_profit >= symbol.ask:
                close = True
            else:
                close = False

            if close:
                if(debug):
                    print("EA Debug: Take profit hidden hit, Trying to close position " +
                          str(ticket)+" ...")
                # HandleTradingEnvironment(debug, Retry_Interval);
                closing = await self.close_position(pos)
                # closing = OrderClose(
                #     OrderTicket(), OrderLots(), Bid, Slip*K, Blue);
                if(debug and not closing):
                    print(
                        "EA Debug: Unexpected Error has happened. While closing the order")
                elif(debug and closing):
                    print("EA Debug: Position successfully closed.")
    #  +------------------------------------------------------------------+
    #  * End of Trigger Hidden Take Profit
    #  +------------------------------------------------------------------+

    # -----------------------------------------------------+
    # * (START)  Volatility-Based Stop Loss
    # -----------------------------------------------------+
    def vol_based_stop_loss(self, isVolatilitySwitchOn: bool, fixedStop: int, volATR: float, volMultiplier: float) -> float:
        """
        Volatility-Based Stop Loss
        """
        if not isVolatilitySwitchOn:
            # ! If Volatility Stop Loss not activated. Stop Loss = Fixed Pips Stop Loss
            StopL = fixedStop
        else:
            # ! Stop Loss in Pips (better to send as decicmal points like I did, DON'T CHANGE)
            # StopL = volMultiplier*volATR/(digits*symobol_point)
            StopL = volMultiplier*volATR
        return (StopL)
    # -----------------------------------------------------+
    # * (END) Volatility-Based Stop Loss
    # -----------------------------------------------------+

    # -----------------------------------------------------+
    # * (START)  Volatility-Based Take Profit
    # -----------------------------------------------------+

    def vol_based_take_profit(self, isVolatilitySwitchOn: bool, fixedTP: int, volATR: float, volMultiplier: float):
        """
        Volatility-Based Take Profit
        """
        if not isVolatilitySwitchOn:
            TakeP = fixedTP
            # ! If Volatility Take Profit not activated. Take Profit = Fixed Pips Take Profit
        else:
            # ! Take Profit in Pips (better to send as decicmal points like I did, DON'T CHANGE)
            # TakeP = volMultiplier*volATR/(digits*symobol_point)
            TakeP = volMultiplier*volATR
            # ! Take Profit in Pips

        return(TakeP)

    # -----------------------------------------------------+
    # * (END)  Volatility-Based Take Profit
    # -----------------------------------------------------+

    # -----------------------------------------------------+
    # * (START)  Update Hidden Trailing List
    # -----------------------------------------------------+
    def update_hidden_trailing_list(self):
        """
        This function clears the elements of your HiddenTrailingList if the corresponding positions has been closed
        """
        hidden_trail_list = self.HiddenTrailingList
        exc = self.exchange

        # ! THIS CODE BLOCK HERE NEED'S TO BE TESTED
        x = 0
        while x < len(hidden_trail_list):
            doesPosExist = False
            orderTicketNumber = hidden_trail_list[x][0]

            if orderTicketNumber != 0:
                pos = exc.get_position_with_ticket_id(orderTicketNumber)

                if pos:
                    doesPosExist = True
                    # break
            if not doesPosExist:
                # ! Deletes elements if the order number does not match any current positions
                del hidden_trail_list[x]
                x = 0 if x == 0 else x - 1
                continue
            x += 1

    # -----------------------------------------------------+
    # * (END)  Update Hidden Trailing List
    # -----------------------------------------------------+

    # -----------------------------------------------------+
    # * (START)  Set and Trigger Hidden Trailing
    # -----------------------------------------------------+

    async def set_and_trigger_hidden_trailing(self, debug: bool, symbol: Symbol, ticket_id, trail_in_point: bool, trail_point_price: Union[int, float], min_profit: int, hidden_sl_df: pd.DataFrame, hidden_trl_sl_df: pd.DataFrame, insert_updt_hidden_trl_sl_db: Callable, step=10):
        """
        hidden_sl_df = get_hidden_sl_db() # this function should called db to collect the df
        hidden_trl_sl_df = get_hidden_trl_sl_db() # this function should called db to collect the df
        insert_updt_hidden_trl_sl_db  # this function should called db to insert the df
        # This function does 2 things.
        >>> 1) It sets hidden trailing stops for all positions
        >>> 2) It closes the positions if hidden trailing stops levels are breached
        """
        exc = self.exchange
        pos = exc.get_position_with_ticket_id(ticket_id)

        min_profit = min_profit * symbol.point * symbol.p_value
        step_value = step * symbol.point * symbol.p_value
        hidden_trl_sl_exist = hidden_trl_sl_df[hidden_trl_sl_df['order_id']
                                               == pos.ticket]

        if pos:
            if len(hidden_trl_sl_exist.index) > 0:

                # if hidden_trail_list[x][0] == pos.ticket:
                tl_sl = hidden_trl_sl_exist['hidden_traili_sl'].iloc[0]
                # trail_points = tl_sl * symbol.point * symbol.p_value
                close = False
                if pos.type == POSITION_TYPE.BUY and tl_sl >= symbol.bid:
                    close = True
                elif pos.type == POSITION_TYPE.SELL and tl_sl <= symbol.ask:
                    close = True

                if close:
                    if(debug):
                        print("EA Debug: Trailing Stop hit, Trying to close position " +
                              str(pos.ticket)+" ...")
                    # HandleTradingEnvironment(debug, Retry_Interval);
                    closing = await self.close_position(pos)
                    # closing = OrderClose(
                    #     OrderTicket(), OrderLots(), Bid, Slip*K, Blue);
                    if(debug and not closing):
                        print(
                            "EA Debug: Unexpected Error has happened. While closing the order")
                    elif(debug and closing):
                        print("EA Debug: Position successfully closed.")

                    # break
                else:
                    # ! Step 2: If there are hidden trailing stop records and the position was not closed in Step 1. We update the hidden trailing stop record.
                    if pos.type == ORDER_TYPE.BUY:
                        if trail_in_point:
                            trail_stop_price = symbol.bid - trail_point_price
                            trail_stop_price = symbol.normalize_price(
                                trail_stop_price)
                        else:
                            trail_stop_price = symbol.normalize_price(
                                trail_point_price)

                        # current_profit_pip = exc.calc_order_profit(
                        #     pos.type, symbol.name, pos.volume, pos.price_open, symbol.bid)
                        current_profit_pip = symbol.bid - pos.price_open
                        current_profit_pip = symbol.normalize_price(
                            current_profit_pip)
                        if trail_stop_price > tl_sl+step_value and current_profit_pip >= min_profit:
                            if debug:
                                print(
                                    f"Trailing stop updated for position {str(pos.ticket)} in  {str(trail_stop_price)}:")
                            insert_updt_hidden_trl_sl_db(
                                pos.ticket, trail_stop_price)
                    elif pos.type == ORDER_TYPE.SELL:
                        if trail_in_point:
                            trail_stop_price = symbol.ask + trail_point_price
                            trail_stop_price = symbol.normalize_price(
                                trail_stop_price)
                        else:
                            trail_stop_price = symbol.normalize_price(
                                trail_point_price)
                        # current_profit_pip = exc.calc_order_profit(
                        #     pos.type, symbol.name, pos.volume, pos.price_open, symbol.ask)
                        current_profit_pip = pos.price_open - symbol.ask
                        current_profit_pip = symbol.normalize_price(
                            current_profit_pip)
                        if trail_stop_price < tl_sl-step_value and current_profit_pip >= min_profit:
                            print(
                                f"Trailing stop updated for position {str(pos.ticket)} in  {str(trail_stop_price)}:")
                            insert_updt_hidden_trl_sl_db(
                                pos.ticket, trail_stop_price)

            # ! Step 3: If there are no hidden trailing stop records, add new record.
            else:
                # Slot is empty
                order_stop_loss = pos.sl
                if order_stop_loss == 0:
                    pos_sl_db = hidden_sl_df[hidden_sl_df['order_id']
                                             == pos.ticket]
                    if pos_sl_db.index > 0:
                        order_stop_loss = pos_sl_db['stop_loss'].iloc[0]

                if pos.type == ORDER_TYPE.BUY:
                    if trail_in_point:
                        # ! if the trail is in point
                        trail_stop_price = symbol.bid - trail_point_price
                        trail_stop_price = symbol.normalize_price(
                            trail_point_price)
                    else:
                        # ! if the trail is in price
                        trail_stop_price = symbol.normalize_price(
                            trail_point_price)
                    # current_profit_pip = exc.calc_order_profit(
                    #     pos.type, symbol.name, pos.volume, pos.price_open, symbol.bid)
                    current_profit_pip = symbol.bid - pos.price_open
                    current_profit_pip = symbol.normalize_price(
                        current_profit_pip)
                    # ! as we don't have any prior sl if our min_profit is achived we will set trail_stop_price as our trail stop
                    if trail_stop_price > (order_stop_loss+step_value) and current_profit_pip >= min_profit:
                        # ! inserting the trail stop loss to db
                        print(
                            f"Trailing stop activated for position {str(pos.ticket)} in  {str(trail_stop_price)}:")
                        insert_updt_hidden_trl_sl_db(
                            pos.ticket, trail_stop_price)

                elif pos.type == ORDER_TYPE.SELL:
                    if trail_in_point:
                        trail_stop_price = symbol.ask + trail_point_price
                        trail_stop_price = symbol.normalize_price(
                            trail_point_price)
                    else:
                        trail_stop_price = symbol.normalize_price(
                            trail_point_price)
                    # current_profit_pip = exc.calc_order_profit(
                    #     pos.type, symbol.name, pos.volume, pos.price_open, symbol.ask)
                    current_profit_pip = pos.price_open - symbol.ask
                    current_profit_pip = symbol.normalize_price(
                        current_profit_pip)
                    # ! as we don't have any prior sl if our min_profit is achived we will set trail_stop_price as our trail stop
                    if trail_stop_price < (order_stop_loss-step_value or order_stop_loss == 0) and current_profit_pip >= min_profit:
                        print(
                            f"Trailing stop activated for position {str(pos.ticket)} in  {str(trail_stop_price)}:")
                        insert_updt_hidden_trl_sl_db(
                            pos.ticket, trail_stop_price)

        else:
            print(f"Position already closed for ticket: {ticket_id}")
    # -----------------------------------------------------+
    # * (END)  Set and Trigger Hidden Trailing
    # -----------------------------------------------------+

    # -----------------------------------------------------+
    # * (START) Update And Review Vol Trailing List (NO NEED FOR THIS SECTION AS IT'S HANDLED IN trailing stop point)
    # -----------------------------------------------------+

    # def update_vol_trailing_list(self, vol_trail_multp: int, min_profit: int, step: int = 10):
    #     """
    #     Updates The List of vol trailing list, with a volatility
    #     """
    #     magic = self.magic_number
    #     vol_trail_list = self.VolTrailingList
    #     exc = self.exchange

    #     # ! THIS CODE BLOCK HERE NEED'S TO BE TESTED
    #     x = 0
    #     while x < len(vol_trail_list):
    #         doesPosExist = False
    #         orderTicketNumber = vol_trail_list[x][ 0]

    #         if orderTicketNumber != 0:
    #             pos = exc.get_position_with_ticket_id(orderTicketNumber)

    #             if pos:
    #                 symbol = Symbol(pos.symbol, TIMEFRAME_M1,terminal=self.TERMINAL)
    #                 doesPosExist = True
    #                 vol_trail_point = vol_trail_multp*vol_trail_list[x][ 1] / \
    #                     (symbol.p_value *
    #                      symbol.point)  # Volatility trailing stop amount in Pips
    #                 order_type = pos.type
    #                 magic_number = pos.magic
    #                 if(magic_number == magic and (order_type == ORDER_TYPE.BUY or order_type == ORDER_TYPE.SELL)):
    #                     self.set_trailing_stop_by_point(
    #                         pos.ticket, vol_trail_point, min_profit, step)

    #                 # break
    #         if not doesPosExist:
    #             # ! Deletes elements if the order number does not match any current positions
    #             vol_trail_list[x][ 0] = 0
    #             vol_trail_list[x][ 1] = 0

    # -----------------------------------------------------+
    # * (END) Update Vol Trailing List
    # -----------------------------------------------------+

    # -----------------------------------------------------+
    # * (START)  Set Volatility Trailing Stop
    # -----------------------------------------------------+

    # def set_vol_trailing_stop(self, volATR: float, vol_trail_multp,  min_profit: int, step: int = 10):
    #     """
    #     set trailing stop for all open order with points
    #     """
    #     vol_trail_list = self.VolTrailingList
    #     exc = self.exchange
    #     magic = self.magic_number

    #     IsVolTrailingStopAdded = False
    #     total_positions = exc.get_position_with_magic(magic)
    #     for position in total_positions:
    #         symbol = Symbol(position.symbol, TIMEFRAME_M1,terminal=self.TERMINAL)
    #         vol_trail_point = vol_trail_multp*volATR / \
    #             (symbol.p_value*symbol.point)  # Volatility trailing stop amount in Pips
    #         order_type = position.type
    #         magic_number = position.magic
    #         if(magic_number == magic and (order_type == ORDER_TYPE.BUY or order_type == ORDER_TYPE.SELL)):
    #             self.set_trailing_stop_by_point(
    #                 position.ticket, vol_trail_point, min_profit, step)
    #             IsVolTrailingStopAdded = True

    #         if IsVolTrailingStopAdded:
    #             for z in range(len(vol_trail_list)+1):
    #                 if vol_trail_list[z][ 0] == 0:
    #                     #  Checks if the element is empty
    #                     vol_trail_list[z][ 0] == position.ticket
    #                     # ! we are only keeping the volATR which was at the time of opening the trdae
    #                     vol_trail_list[z][ 1] == volATR / \
    #                         (symbol.p_value *
    #                          symbol.point)  # Volatility trailing stop amount in Pips
    #                     # ! break from this loop only
    #                     break

    # -----------------------------------------------------+
    # * (END)  Set Volatility Trailing Stop
    # -----------------------------------------------------+

    # +------------------------------------------------------------------+
    #  * (START) Update Hidden Breakeven Stops List
    # +------------------------------------------------------------------+
    def update_hidden_BE_list(self):
        """
        This function clears the elements of your HiddenBEList if the corresponding positions has been closed
        """
        hidden_be_list = self.HiddenBEList
        exc = self.exc
        x = 0
        while x < len(hidden_be_list):
            doesPosExist = False
            orderTicketNumber = hidden_be_list[x]
            if(orderTicketNumber != 0):
                #  Order exists
                pos = exc.get_position_with_ticket_id(
                    orderTicketNumber)

                if pos:
                    doesPosExist = True
                    break
            if(doesPosExist == False):
                # ! Deletes elements if the order number does not match any current positions
                del hidden_be_list[x]
                x = 0 if x == 0 else x - 1
                continue

            x += 1

    #  +------------------------------------------------------------------+
    #  * (End) Update Hidden Breakeven Stops List
    #  +------------------------------------------------------------------+

    # +------------------------------------------------------------------+
    #  * (START) Set and Trigger Hidden Breakeven Stops
    # +------------------------------------------------------------------+
    async def set_and_trigger_BE_hidden(self, debug: bool, symbol: Symbol, ticket: int, min_profit_pip: int, lock_profit: int, be_hdn_df: pd.DataFrame, hidden_sl_df: pd.DataFrame, insert_hidden_be_sl_db: Callable):
        """
        hidden_sl_df = get_hidden_sl_db() # this function should called db to collect the df
        be_hdn_df = get_hidden_be_db() # this function should called db to collect the df
        insert_hidden_be_sl_db  # this function should called db to insert the df
        This function scans through the current positions and does 2 things:
        1) If the position is in the hidden breakeven list, it closes it if the appropriate conditions are met
        2) If the positon != the hidden breakeven list, it adds it to the list if the appropriate conditions are met
        """
        exc = self.exchange
        pos = exc.get_position_with_ticket_id(ticket)
        be_hdn_df = be_hdn_df[be_hdn_df['order_id'] == ticket]
        if pos:
            if len(be_hdn_df.index) != 0:
                be_hidden = be_hdn_df['break_even_stop'].iloc[0]
                if pos.type == POSITION_TYPE.BUY and be_hidden <= symbol.bid:
                    # ! price came down and bid is equal to price open so close the position
                    close = True
                elif pos.type == POSITION_TYPE.SELL and be_hidden >= symbol.ask:
                    # ! price gone up and ask is equal to price open so close the position
                    close = True

                if close:
                    if(debug):
                        print("EA Debug:Break Even Stop hit, Trying to close position " +
                              str(pos.ticket)+" ...")
                    # HandleTradingEnvironment(debug, Retry_Interval);
                    closing = await self.close_position(pos)
                    # closing = OrderClose(
                    #     OrderTicket(), OrderLots(), Bid, Slip*K, Blue);
                    if(debug and not closing):
                        print(
                            "EA Debug: Unexpected Error has happened. While closing the order")
                    elif(debug and closing):
                        print("EA Debug: Position successfully closed.")
            else:
                min_profit = min_profit_pip * symbol.point * symbol.p_value

                lock_profit_val = 0
                if lock_profit > 0:
                    lock_profit_val = lock_profit * symbol.point * symbol.p_value

                order_stop_loss = pos.sl
                if order_stop_loss == 0:
                    pos_sl_db = hidden_sl_df[hidden_sl_df['order_id']
                                             == pos.ticket]
                    order_stop_loss = pos_sl_db['stop_loss']

                if pos.type == POSITION_TYPE.BUY:
                    break_even_stop = pos.price_open + lock_profit_val
                    if break_even_stop > order_stop_loss and (symbol.bid - pos.price_open) > min_profit:
                        add_be = True
                elif pos.type == POSITION_TYPE.SELL:
                    break_even_stop = pos.price_open - lock_profit_val
                    if break_even_stop > order_stop_loss and (pos.price_open - symbol.ask) > min_profit:
                        add_be = True
                else:
                    add_be = False

                if add_be:
                    insert_hidden_be_sl_db(pos.ticket, break_even_stop)
                    if(debug):
                        print("EA debug: Order "+str(pos.ticket) +
                              " assigned with a hidden breakeven stop.")
        else:
            print(f"Position already closed for ticket: ", ticket)
    #  +------------------------------------------------------------------+
    #  * (END) Set and Trigger Hidden Breakeven Stops
    #  +------------------------------------------------------------------+

    # -----------------------------------------------------+
    # * (START)  Update Hidden Volatility Trailing List
    # -----------------------------------------------------+

    def update_hidden_vol_trailing_list(self):
        """
        This function clears the elements of your HiddenVolTrailingList if the corresponding positions has been closed
        """
        hidden_vol_trail_list = self.HiddenVolTrailingList
        exc = self.exchange

        # ! THIS CODE BLOCK HERE NEED'S TO BE TESTED
        x = 0
        while x < len(hidden_vol_trail_list):
            doesPosExist = False
            orderTicketNumber = hidden_vol_trail_list[x][0]

            if orderTicketNumber != 0:
                pos = exc.get_position_with_ticket_id(orderTicketNumber)

                if pos:
                    doesPosExist = True
                    # break
            if not doesPosExist:
                # ! Deletes elements if the order number does not match any current positions
                del hidden_vol_trail_list[x]
                x = 0 if x == 0 else x - 1
                continue
            x += 1

    # -----------------------------------------------------+
    # * (END) Update Hidden Volatility Trailing List
    # -----------------------------------------------------+

    # -----------------------------------------------------+
    # * (START)  Set Trigger Review Hidden Vol Trailing
    # -----------------------------------------------------+

    async def set_trigger_review_hidden_vol_trailing(self, debug: bool, symbol: Symbol, volATR: float, vol_trail_multp: int,  min_profit: int, step: int = 10):
        """
        # This function does 2 things.
        >>> 1) It sets hidden trailing stops for all positions
        >>> 2) It closes the positions if hidden trailing stops levels are breached
        """
        magic = self.magic_number
        doesHiddenVolTrailingRecordExist = False
        hidden_Vol_trail_list = self.HiddenVolTrailingList
        exc = self.exchange
        positions = exc.get_position_with_magic(magic)
        symbol_pos = list(filter(lambda p: p.symbol == symbol.name, positions))

        for pos in symbol_pos:
            x = 0
            while x < len(hidden_Vol_trail_list):
                vol_trailing_stop_level = 0
                # Volatility trailing stop amount in Pips
                vol_trail_point = vol_trail_multp * volATR
                min_profit = min_profit * symbol.point * symbol.p_value
                step_value = step * symbol.point * symbol.p_value
                if hidden_Vol_trail_list[x][0] == pos.ticket:
                    doesHiddenTrailingRecordExist = True
                    vol_trailing_stop_level = hidden_Vol_trail_list[x][1]
                    close = False

                    # ! TRIGGER
                    if doesHiddenTrailingRecordExist:
                        if pos.type == POSITION_TYPE.BUY and vol_trailing_stop_level >= symbol.bid:
                            close = True
                        elif pos.type == POSITION_TYPE.SELL and vol_trailing_stop_level <= symbol.ask:
                            close = True

                        if close:
                            if(debug):
                                print("EA Debug:Hidden vol trail hit, Trying to close position " +
                                      str(pos.ticket)+" ...")
                            # HandleTradingEnvironment(debug, Retry_Interval);
                            closing = await self.close_position(pos)
                            # closing = OrderClose(
                            #     OrderTicket(), OrderLots(), Bid, Slip*K, Blue);
                            if(debug and not closing):
                                print(
                                    "EA Debug: Unexpected Error has happened. While closing the order")
                            elif(debug and closing):
                                print("EA Debug: Position successfully closed.")

                            # break
                        # ! REVIEW
                        else:
                            # ! Step 2: If there are hidden trailing stop records and the position was not closed in Step 1. We update the hidden trailing stop record.
                            if pos.type == ORDER_TYPE.BUY:
                                trail_stop_price = symbol.bid - vol_trail_point
                                trail_stop_price = symbol.normalize_price(
                                    trail_stop_price)
                                # current_profit_pip = exc.calc_order_profit(
                                #     pos.type, symbol.name, pos.volume, pos.price_open, symbol.bid)
                                current_profit_pip = symbol.bid - pos.price_open
                                current_profit_pip = symbol.normalize_price(
                                    current_profit_pip)
                                if trail_stop_price > vol_trailing_stop_level+step_value and current_profit_pip >= min_profit:
                                    hidden_Vol_trail_list[x][1] = trail_stop_price
                            elif pos.type == ORDER_TYPE.SELL:
                                trail_stop_price = symbol.ask + vol_trail_point
                                trail_stop_price = symbol.normalize_price(
                                    trail_stop_price)
                                # current_profit_pip = exc.calc_order_profit(
                                #     pos.type, symbol.name, pos.volume, pos.price_open, symbol.ask)
                                current_profit_pip = pos.price_open - symbol.ask
                                current_profit_pip = symbol.normalize_price(
                                    current_profit_pip)
                                if trail_stop_price < vol_trailing_stop_level-step_value and current_profit_pip >= min_profit:
                                    hidden_Vol_trail_list[x][1] = trail_stop_price
                # ! (SET) If there are no hidden trailing stop records, add new record.
            if not doesHiddenVolTrailingRecordExist:
                order_stop_loss = pos.sl
                if order_stop_loss == 0:
                    hidden_sl_list = self.HiddenSLList
                    x = 0
                    while x < len(hidden_sl_list):
                        #  Looping through all order number in list
                        # doesOrderExist = False
                        orderTicketNumber = hidden_sl_list[x][0]
                        orderSL = hidden_sl_list[x][1]
                        if orderTicketNumber == pos.ticket:
                            if pos.type == ORDER_TYPE.BUY:
                                order_stop_loss = pos.price_open - \
                                    (orderSL * symbol.p_value * symbol.point)
                                break
                            elif pos.type == ORDER_TYPE.SELL:
                                order_stop_loss = pos.price_open + \
                                    (orderSL * symbol.p_value * symbol.point)
                                break

                        x += 1

                if pos.type == ORDER_TYPE.BUY:
                    vol_trail_point = vol_trail_multp*volATR
                    trail_stop_price = symbol.bid - vol_trail_point
                    trail_stop_price = symbol.normalize_price(
                        trail_stop_price)
                    # current_profit_pip = exc.calc_order_profit(
                    #     pos.type, symbol.name, pos.volume, pos.price_open, symbol.bid)
                    current_profit_pip = symbol.bid - pos.price_open
                    current_profit_pip = symbol.normalize_price(
                        current_profit_pip)

                    # ! as we don't have any prior sl if our min_profit is achived we will set trail_stop_price as our trail stop
                    if trail_stop_price > (order_stop_loss+step) and current_profit_pip >= min_profit:
                        hidden_Vol_trail_list.append(
                            [pos.ticket, trail_stop_price])

                elif pos.type == ORDER_TYPE.SELL:
                    vol_trail_point = vol_trail_multp*volATR
                    trail_stop_price = symbol.ask + vol_trail_point
                    trail_stop_price = symbol.normalize_price(
                        trail_stop_price)
                    # current_profit_pip = exc.calc_order_profit(
                    #     pos.type, symbol.name, pos.volume, pos.price_open, symbol.ask)
                    current_profit_pip = pos.price_open - symbol.ask
                    current_profit_pip = symbol.normalize_price(
                        current_profit_pip)
                    # ! as we don't have any prior sl if our min_profit is achived we will set trail_stop_price as our trail stop
                    if trail_stop_price < (order_stop_loss-step or order_stop_loss == 0) and current_profit_pip >= min_profit:
                        hidden_Vol_trail_list.append(
                            [pos.ticket, trail_stop_price])

    # -----------------------------------------------------+
    # * (END)  Set Trigger Review Hidden Vol Trailing
    # -----------------------------------------------------+

    # -----------------------------------------------------+
    # * (START) OPEN FROM MARKET
    # -----------------------------------------------------+

    async def OpenPositionMarket(self, symbol: Symbol, lot, order_type: int, SL: Union[int, None], TP: Union[int, None], Journaling: bool, Max_Retries_Per_Tick: int, ECN: bool = False, sl_in_pip: bool = True, tp_in_pip: bool = True):
        """
        This function submits new orders
        >>> `oreder_type`: POSITION_TYPE.BUY,POSITION_TYPE.SELL
        >>> `lot`: lot is volume for order
        >>> `SL`: stop loss in point
        >>> `TP`: take profit in point
        >>> `Journaling`: To print information
        >>> `P`: P value for pair
        >>> `Max_Retries_Per_Tick`: max tries on error
        >>> `ECN`: If broker is a ECN broker, order will be placed first, than we will modify with SL & TP
        """
        exc = self.exchange
        magic = self.magic_number
        Slip = self.slipaage
        P = symbol.p_value

        # ! this is a temporary code remove it in the next run
        # lot = lot if lot > 0 else 0.01
        # -----------------------------------------------------+
        # * (START)  check margin available
        # -----------------------------------------------------+
        is_margin_available = exc.validate_order_margin(
            order_type, symbol.name, lot, symbol.ask)

        if not is_margin_available:
            print("Can not open a trade. Not enough free margin to open " +
                  str(lot)+" on "+symbol.name)
            return dict()

        # -----------------------------------------------------+
        # * (END) check margin available
        # -----------------------------------------------------+
        if order_type == ORDER_TYPE.BUY or order_type == ORDER_TYPE.SELL:
            # ! execute only if we have buy or sell order type

            # ! Slippage is in points. 1 point = 0.0001 on 4 digit broker and 0.00001 on a 5 digit broker
            slippage = Slip*P
            buy_sell = 'BUY ' if order_type == ORDER_TYPE.BUY else 'SELL '
            comment = " "+buy_sell+"(#"+str(magic)+")"
            # ! if this order doesn't get executed under 5 seconds it will be expired(FOR MARKET ORDER EXPIRATION NOT NEEDED)
            # expiration = datetime.today() + timedelta(minutes=5)
            tries = 0
            stoploss = SL
            takeprofit = TP
            price = 0
            # if not ECN:
            #     # ! Edits stops and take profits before the market order is placed
            #     while(tries < Max_Retries_Per_Tick):
            #         # -----------------------------------------------------+
            #         # * (START)  Retriving Bid Ask Price
            #         # -----------------------------------------------------+
            #         bid = symbol.bid
            #         ask = symbol.ask
            #         # -----------------------------------------------------+
            #         # * (END) Retriving Bid Ask Price
            #         # -----------------------------------------------------+

            #         if(order_type == ORDER_TYPE.BUY):
            #             price = ask
            #         elif(order_type == ORDER_TYPE.SELL):
            #             price = bid

            #         # ! Sets Take Profits and Stop Loss. Check against Stop Level Limitations.
            #         if isinstance(SL, int):
            #             # ! stop defined in pips
            #             if(order_type == ORDER_TYPE.BUY and SL != 0):
            #                 stoploss = symbol.normalize_price(
            #                     ask-SL*P*symbol.point)
            #                 stoploss = self.adjust_below_stop_level(
            #                     symbol, stoploss)

            #             elif(order_type == ORDER_TYPE.SELL and SL != 0):
            #                 stoploss = symbol.normalize_price(bid+SL*P*symbol.point,
            #                                                   symbol.digits)
            #                 stoploss = self.adjust_above_stop_level(
            #                     symbol, stoploss)
            #         elif isinstance(SL, float):
            #             # ! Stop loss defined with exact price
            #             if(order_type == ORDER_TYPE.BUY and SL != 0.00):
            #                 # stoploss = symbol.normalize_price(
            #                 #     ask-SL)
            #                 stoploss = self.adjust_below_stop_level(
            #                     symbol, SL)

            #             elif(order_type == ORDER_TYPE.SELL and SL != 0):
            #                 # stoploss = symbol.normalize_price(bid+SL)
            #                 stoploss = self.adjust_above_stop_level(
            #                     symbol, SL)
            #         if isinstance(TP, int):
            #             # ! TAKE PROFIT defined in pips
            #             if(order_type == ORDER_TYPE.BUY and TP != 0):
            #                 takeprofit = symbol.normalize_price(
            #                     ask+TP*P*symbol.point)
            #                 takeprofit = self.adjust_above_stop_level(
            #                     symbol, takeprofit)

            #             elif(order_type == ORDER_TYPE.SELL and TP != 0):
            #                 takeprofit = symbol.normalize_price(
            #                     bid-TP*P*symbol.point)
            #                 takeprofit = self.adjust_below_stop_level(
            #                     symbol, takeprofit)
            #         elif isinstance(TP, float):
            #             # ! TAKE PROFIT DEFINED IN PRICE
            #             if(order_type == ORDER_TYPE.BUY and TP != 0):
            #                 takeprofit = self.adjust_above_stop_level(
            #                     symbol, TP)
            #             elif(order_type == ORDER_TYPE.SELL and TP != 0):
            #                 takeprofit = self.adjust_below_stop_level(
            #                     symbol, TP)

            #         if(Journaling):
            #             print("EA Journaling: Trying to place a market order...")
            #         # HandleTradingEnvironment(Journaling, Retry_Interval);
            #         try:
            #             result = exc.market_order_send(symbol=symbol.name, price=price, volume=lot, stop_loss=stoploss, take_profit=takeprofit,
            #                                            order_type=order_type, slippage=slippage, magic=magic, comment=comment, expiration=None)
            #             # result = exc.market_order_send(symbol=symbol.name, price=price, volume=lot,
            #             #                                order_type=order_type, slippage=slippage, magic=magic, comment=comment)

            #             assert result.comment != MARKET_CLOSED, 'Market is closed trade at a later time'
            #             pos = exc.get_position_from_order_send_result(result)
            #             return pos
            #         except Exception as ex:
            #             print(ex.args)

            #         tries += tries

            # if(ECN):  # ! Edits stops and take profits after the market order is placed

            if(order_type == ORDER_TYPE.BUY):
                price = symbol.ask
            elif(order_type == ORDER_TYPE.SELL):
                price = symbol.bid

            if(Journaling):
                timezone = pytz.timezone('EET')

                print("EA Journaling: Trying to place a market order at ... ",
                      datetime.now(tz=timezone).time())
            pos = dict()
            try:
                result = await exc.market_order_send(symbol=symbol.name, volume=lot, price=price,
                                                     order_type=order_type, slippage=slippage, magic=magic, comment=comment, expiration=None)

                assert result.comment != MARKET_CLOSED, 'Market is closed trade at a later time'
                pos = exc.get_position_from_order_send_result(result)
                print("Order successfully placed at ",
                      pd.to_datetime(pos.time, unit='s'))
            except Exception as ex:
                print(ex.args)

            if(pos and ((SL != 0 and SL != None) or (TP != 0 and TP != None))):
                # ! Sets Take Profits and Stop Loss. Check against Stop Level Limitations.
                if sl_in_pip:
                    if(order_type == ORDER_TYPE.BUY and SL != 0):
                        stoploss = symbol.normalize_price(
                            min(pos.price_open, symbol.bid)-SL * P * symbol.point)
                        stoploss = self.adjust_below_stop_level(
                            symbol, stoploss)

                    elif(order_type == ORDER_TYPE.SELL and SL != 0):

                        stoploss = symbol.normalize_price(
                            max(pos.price_open, symbol.ask)+(SL * P * symbol.point))
                        stoploss = self.adjust_above_stop_level(
                            symbol, stoploss)
                if tp_in_pip:
                    if(order_type == ORDER_TYPE.BUY and (TP != 0 and TP != None)):
                        takeprofit = symbol.normalize_price(
                            max(pos.price_open, symbol.ask)+TP * P * symbol.point)
                        takeprofit = self.adjust_above_stop_level(
                            symbol, takeprofit)

                    elif(order_type == ORDER_TYPE.SELL and (TP != 0 and TP != None)):
                        takeprofit = symbol.normalize_price(
                            min(pos.price_open, symbol.bid)-TP * P * symbol.point)
                        takeprofit = self.adjust_below_stop_level(
                            symbol, takeprofit)

                ModifyOpen = dict()
                while(not ModifyOpen) and tries <= Max_Retries_Per_Tick:
                    try:
                        ModifyOpen = exc.modify_sl_tp(
                            takeprofit, stoploss, pos)
                        return exc.get_position_with_ticket_id(pos.ticket)
                    except Exception as ex:
                        print(ex.args)
                    tries += 1
                return exc.get_position_with_ticket_id(pos.ticket)

            if pos:
                return pos

        return dict()

    # -----------------------------------------------------+
    # * (END) OPEN FROM MARKET
    # -----------------------------------------------------+

    # -----------------------------------------------------+
    # * (START) OPEN PENDING ORDERS
    # -----------------------------------------------------+
    async def OpenPendingPosition(self, symbol: Symbol, lot, open_price: float, order_type: Entry, SL: Union[int, None], TP: Union[int, None], Journaling: bool, Max_Retries_Per_Tick: int, ECN: bool = False, sl_in_pip: bool = True, tp_in_pip: bool = True, expiration=None):
        """
        This function submits new orders
        >>> `oreder_type`: POSITION_TYPE.BUY,POSITION_TYPE.SELL
        >>> `lot`: lot is volume for order
        >>> `SL`: stop loss in point
        >>> `TP`: take profit in point
        >>> `Journaling`: To print information
        >>> `P`: P value for pair
        >>> `Max_Retries_Per_Tick`: max tries on error
        >>> `ECN`: If broker is a ECN broker, order will be placed first, than we will modify with SL & TP
        >>> `Expiration`: for how long the order will be live, in hours (CAN'T USE IT YET)
        """
        exc = self.exchange
        magic = self.magic_number
        P = symbol.p_value
        slippage = self.slipaage * P

        # -----------------------------------------------------+
        # * (START)  Trade Info
        # -----------------------------------------------------+
        account_info = exc.get_account_info()
        tries = 0
        # expiration = datetime.today() + expiration
        stoploss = SL
        takeprofit = TP
        price = 0
        # -----------------------------------------------------+
        # * (END)  Trade Info
        # -----------------------------------------------------+

        # -----------------------------------------------------+
        # * (START)  Retriving Bid Ask Price
        # -----------------------------------------------------+
        if(order_type == Entry.Buy):
            price = symbol.ask
            if open_price > price:
                order_type = ORDER_TYPE.BUY_STOP
            elif open_price < price:
                order_type = ORDER_TYPE.BUY_LIMIT
            else:
                order_type = ORDER_TYPE.BUY
        elif(order_type == Entry.Sell):
            price = symbol.bid
            if open_price > price:
                order_type = ORDER_TYPE.SELL_LIMIT
            elif open_price < price:
                order_type = ORDER_TYPE.SELL_STOP
            else:
                order_type = ORDER_TYPE.SELL

        # -----------------------------------------------------+
        # * (END) Retriving Bid Ask Price
        # -----------------------------------------------------+
        com_buy = order_type.name
        comment = ""+com_buy+"(#"+str(magic)+")"

        # -----------------------------------------------------+
        # * (START)  check margin available
        # -----------------------------------------------------+

        is_margin_available = exc.validate_order_margin(
            order_type, symbol, lot, open_price)

        if not is_margin_available:
            print("Can not open a trade. Not enough free margin to open " +
                  lot+" on "+symbol)
            return dict()
        # -----------------------------------------------------+
        # * (END) check margin available
        # -----------------------------------------------------+

        # -----------------------------------------------------+
        # * (START)  STOP LOSS Take Profit
        # -----------------------------------------------------+

        # ! Sets Take Profits and Stop Loss. Check against Stop Level Limitations.
        if((SL != 0 and SL != None) or (TP != 0 and TP != None)):
            if sl_in_pip:
                if order_type == ORDER_TYPE.BUY_LIMIT or order_type == ORDER_TYPE.BUY_STOP and SL != 0:
                    stoploss = symbol.normalize_price(
                        open_price - (SL * P * symbol.point))
                    stoploss = self.adjust_below_stop_level(
                        symbol, stoploss)
                elif order_type == ORDER_TYPE.SELL_LIMIT or order_type == ORDER_TYPE.SELL_STOP and SL != 0:
                    stoploss = symbol.normalize_price(
                        open_price + (SL * P * symbol.point))
                    stoploss = self.adjust_above_stop_level(
                        symbol, stoploss)
            if tp_in_pip:
                if order_type == ORDER_TYPE.BUY_LIMIT or order_type == ORDER_TYPE.BUY_STOP and TP != 0:
                    takeprofit = symbol.normalize_price(
                        open_price + (TP * P * symbol.point))
                    takeprofit = self.adjust_above_stop_level(
                        symbol, takeprofit)
                if order_type == ORDER_TYPE.SELL_LIMIT or order_type == ORDER_TYPE.SELL_STOP and TP != 0:
                    takeprofit = symbol.normalize_price(
                        open_price - (TP * P * symbol.point))
                    takeprofit = self.adjust_below_stop_level(
                        symbol, takeprofit)

        # -----------------------------------------------------+
        # * (END)  STOP LOSS Take Profit
        # -----------------------------------------------------+

        if(Journaling):
            timezone = pytz.timezone('EET')

            print(f"EA Journaling: Trying to place a pending {order_type.name} order at ... ",
                  datetime.now(tz=timezone).time())

        pos = dict()
        try:
            # ! Note: We did not modify Open Price if it breaches the Stop Level Limitations as Open Prices are sensitive and important. It is unsafe to change it automatically.
            result = await exc.market_order_send(symbol=symbol.name, action=TRADE_ACTION.PENDING, price=open_price, volume=lot, stop_loss=stoploss, take_profit=takeprofit, order_type=order_type, slippage=slippage, magic=magic, comment=comment, type_time=ORDER_TIME.GTC, expiration=expiration)

            assert result.comment != MARKET_CLOSED, 'Market is closed trade at a later time'
            ticket = result.order
            all_pen = exc.get_all_pending_position()
            pos = filter(lambda position: position.ticket == ticket
                         and position.symbol == symbol.name, all_pen)
            pos = list(pos)[0]
            print("Order successfully placed at ",
                  pd.to_datetime(pos.time_setup, unit='s'))
        except Exception as ex:
            print(ex.args)

        return pos

    # -----------------------------------------------------+
    # * (END) OPEN PENDING ORDERS
    # -----------------------------------------------------+

    # -----------------------------------------------------+
    # * (START)  CLOSE/DELETE ORDERS AND POSITIONS
    # -----------------------------------------------------+

    # -----------------------------------------------------+
    # * (END)
    # -----------------------------------------------------+


def main():
    symbol = mt5.symbol_info('GBPUSD')
    pos = sell(symbol, volume=0.01)
    modify_sl(pos.identifier, 300)
    modify_tp(pos.identifier, 200)
    # assert pos and pos.sl and pos.tp  # and pos.volume == 1.0

    time.sleep(2)
    r = close_position(pos)
    # ! checking if the retcode is done
    assert r.retcode == mt5.TRADE_RETCODE.DONE
    # ! checking if any other open position for this symbol
    assert not mt5.positions_get(symbol=symbol)


if __name__ == '__main__':
    conn = connected()
    with conn:
        ticks = mt5.copy_rates('GBPUSD', mt5.const.TIMEFRAME_H1)
        # create DataFrame out of the obtained data
        ohlcv = pd.DataFrame(ticks)
        # convert time in seconds into the datetime format
        ohlcv['time'] = pd.to_datetime(ohlcv['time'], unit='s')
        ohlcv.set_index('time', drop=True, inplace=True)
        spreads_df = ohlcv['2020-01-01':'2020-10-20'].sort_values(
            'spread', ascending=False)
        main()
        # deal = mt5.history_deals_get()
        # pos = mt5.positions_get(ticket=deal.position_id)[0]

    # trade = Trade('GBPUSD', magic=23458)



import pymt5adapter as mt5
import logging


class Connection(object):
    def __init__(self,
                 path,
                 portable,
                 server,
                 login,
                 password,
                 timeout,
                 logger,  # default is None
                 ensure_trade_enabled,  # default is False
                 enable_real_trading,  # default is False
                 raise_on_errors,  # default is False
                 return_as_dict,  # default is False
                 return_as_native_python_objects   # default is False)
                 ):
        """
        connection base class for all trading related work
        """
        self.connected = None
        self.connected = self.connect(path=path,
                                      portable=portable,
                                      server=server,
                                      login=login,
                                      password=password,
                                      timeout=timeout,
                                      logger=logger,  # default is None
                                      ensure_trade_enabled=ensure_trade_enabled,  # default is False
                                      enable_real_trading=enable_real_trading,  # default is False
                                      raise_on_errors=raise_on_errors,  # default is False
                                      return_as_dict=return_as_dict,  # default is False
                                      return_as_native_python_objects=return_as_native_python_objects,  # default is False
                                      )

    def connect(
        self,
        path,
        portable=False,
        server='XMGlobal-MT5 2',
        login=64158516,
        password='qM0PyD9sjsIg',
        timeout=5000,
        logger=None,  # default is None
        ensure_trade_enabled=True,  # default is False
        enable_real_trading=False,  # default is False
        raise_on_errors=True,  # default is False
        return_as_dict=False,  # default is False
        return_as_native_python_objects=False  # default is False
    ):
        """
        function to get a connection to mt5
        """

        # LOGPATH = './my_mt5_log.log'
        # ! If you need logger uncomment below code, but logger takes too much space, and it isn't that useful to me still, that's why I commented out below code.
        # logger = mt5.get_logger(loglevel=logging.DEBUG,
        #                         path_to_logfile=logger)

        mt5_connected = mt5.connected(
            # path=r'C:\Program Files\XM Global MT5\terminal64.exe',
            path=path,
            portable=portable,
            server=server,
            login=login,
            password=password,
            timeout=timeout,
            logger=logger,  # default is None
            ensure_trade_enabled=ensure_trade_enabled,  # default is False
            enable_real_trading=enable_real_trading,  # default is False
            raise_on_errors=raise_on_errors,  # default is False
            return_as_dict=return_as_dict,  # default is False
            return_as_native_python_objects=return_as_native_python_objects,  # default is False
        )
        return mt5_connected



import asyncio
from MT5TrdHelper.helpers.entries import Entry
import time
from typing import Tuple
import pandas as pd
from datetime import datetime
from localconfig import config
from MT5TrdHelper.connect.connection import Connection
from MetaTrader5 import TIMEFRAME_H1, TIMEFRAME_H4, TIMEFRAME_M1
from pymt5adapter import const
from pymt5adapter.context import _ContextAwareBase
from pymt5adapter.core import copy_rates_from_pos, copy_ticks_from, copy_rates, copy_rates_from
from pymt5adapter.core import symbol_info as sym_info
from pymt5adapter.core import symbol_info_tick
from pymt5adapter.core import symbol_select
from pymt5adapter.types import SymbolInfo
from pymt5adapter.types import Union

connect_path = "D:\\DATA_SET_MACHINE_LEARNING\\trading_engine\\my_trading_system\\connect\\connection.ini"


def make_ohlcv(array_of_price):
    """
    With array of price from MT5 it creates ohlcv data
    """
    # create DataFrame out of the obtained data
    ohlcv = pd.DataFrame(array_of_price)
    # convert time in seconds into the datetime format
    ohlcv['time'] = pd.to_datetime(ohlcv['time'], unit='s')
    ohlcv.set_index('time', drop=False, inplace=True)
    ohlcv.columns = ['time', 'Open', 'High', 'Low',
                             'Close', 'Volume', 'Spread', 'Real_Volume']

    return ohlcv


class Symbol(_ContextAwareBase, Connection):
    calc_timeframe = 0
    current_bar = 0
    last_bar = 0
    first_time = True

    current_tick = 0
    first_tick = True
    last_tick = 0

    def __init__(self, symbol: Union[str, SymbolInfo], timeframe, max_bars: int = None, terminal='five_prcnt_real_terminal_connection'):
        # -----------------------------------------------------+
        # * (START)  reading all the configuration and initializing connection class
        # -----------------------------------------------------+
        # config.read('exchange\\connection.ini')
        # ! for interactive python shell
        config.read(connect_path)
        # if terminal == 'GLOBAL':
        #     keyargs = dict(list(config.real_terminal_connection))
        # elif terminal == 'XM':
        #     keyargs = dict(list(config.xm_terminal_connection))
        # super(Symbol, self).__init__(**keyargs)
        # -----------------------------------------------------+
        # * (END)  reading all the configuration and initializing connection class
        # -----------------------------------------------------+
        self.terminal = terminal
        self.name = symbol
        self._timeframe = timeframe
        self.max_bars = max_bars

    @property
    def terminal(self):
        return self._terminal

    @terminal.setter
    def terminal(self, terminal: str):
        self._terminal = terminal
        keyargs = dict(config.items(terminal))
        super(Symbol, self).__init__(**keyargs)

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, symbol):
        try:
            self._name = symbol.name
            self._info = symbol
        except AttributeError:
            self._name = symbol
            self._info = None
        self._refresh()

    @property
    def df(self) -> pd.DataFrame:
        return self.get_rates_bar(ohlcv=True)

    @property
    def p_value(self) -> int:
        """
        P value for pip calculation 4 digit and 5 digit broker
        """
        digits = self.digits
        if digits == 5 or digits == 3 or digits == 1:
            return 10  # ! to account for 5 digit broker to make the pips for 4 digit
        else:
            return 1

    @property
    def symbol_info(self):
        """
        Get all the information of this symbol, each key is a property of this symbol
        """
        with self.connected:
            symbol_info = sym_info(self.name)
            return symbol_info

    @property
    def timeframe(self):
        return self._timeframe

    @timeframe.setter
    def timeframe(self, timeframe):
        self._timeframe = timeframe

    @property
    def select(self):
        return self._select

    @select.setter
    def select(self, enable: bool):
        with self.connected:
            x = symbol_select(self._name, enable)
            self._select = enable if x else not enable

    @property
    def tick(self) -> Tuple:
        with self.connected:
            self._tick = symbol_info_tick(self.name)
            return self._tick

    @property
    def spread(self) -> float:
        return int(round(self.spread_float / self.trade_tick_size))

    @property
    def bid(self) -> float:
        return self.tick.bid

    @property
    def ask(self) -> float:
        return self.tick.ask

    def _data(self, key):
        rate = self.get_rates_position()
        try:
            return rate[:][key]
        except:
            return 0

    @property
    def tick_volume(self) -> float:
        return self.tick.volume

    @property
    def time(self) -> datetime:
        """
        last bar time for defined timeframe, by default daily
        """
        return pd.to_datetime(self._data('time')[-1], unit='s')

    @property
    def time_bars(self) -> pd.Series:
        """
        last bar time for defined timeframe, by default daily
        """
        df = self.get_rates_bar(ohlcv=True)
        time = pd.Series(df['time'], index=df.index)
        return time

    @property
    def close(self) -> float:
        """
        last close price for defined timeframe, by default daily
        """
        return self._data('close')[-1]

    @property
    def close_bars(self) -> pd.Series:
        """
        all close price for defined timeframe, by default daily
        """
        df = self.get_rates_bar(ohlcv=True)
        Close = pd.Series(df['Close'], index=df.index)
        return Close

    @property
    def low(self) -> float:
        """
        last low price for defined timeframe, by default daily
        """
        return self._data('low')[-1]

    @property
    def low_bars(self) -> pd.Series:
        """
        all low price for defined timeframe, by default daily
        """
        df = self.get_rates_bar(ohlcv=True)
        lows = pd.Series(df['Low'], index=df.index)
        return lows

    @property
    def high(self) -> float:
        """
        last low price for defined timeframe, by default daily
        """
        return self._data('high')[-1]

    @property
    def high_bars(self) -> pd.Series:
        """
        all low price for defined timeframe, by default daily
        """
        df = self.get_rates_bar(ohlcv=True)
        highs = pd.Series(df['High'], index=df.index)
        return highs

    @property
    def open(self) -> float:
        """
        last open price for defined timeframe, by default daily
        """
        return self._data('open')[-1]

    @property
    def open_bars(self) -> pd.Series:
        """
        all open price for defined timeframe, by default daily
        """
        df = self.get_rates_bar(ohlcv=True)
        Open = pd.Series(df['Open'], index=df.index)
        return Open

    @property
    def real_volume(self) -> float:
        """
        last real_volume for defined timeframe, by default daily
        """
        return self._data('real_volume')[-1]

    @property
    def real_volume_bars(self) -> pd.Series:
        """
        all real_volume for defined timeframe, by default daily
        """
        df = self.get_rates_bar(ohlcv=True)
        Real_Volume = pd.Series(df['Real_Volume'], index=df.index)
        return Real_Volume

    @property
    def volume(self) -> float:
        """
        last volume price for defined timeframe, by default daily
        """
        return self._data('volume')[-1]

    @property
    def volume_bars(self) -> pd.Series:
        """
        all volume price for defined timeframe, by default daily
        """
        df = self.get_rates_bar(ohlcv=True)
        Volume = pd.Series(df['Volume'], index=df.index)
        return Volume

    # @property
    # def volume_real(self):
    #     return self.tick.volume_real

    @property
    def is_new_bar(self) -> bool:
        """
        If a new bar occured for defined timeframe
        """
        if self.calc_timeframe != self.timeframe:
            self.calc_timeframe = self.timeframe
            self.first_bar = True

        self.current_bar = self.time

        if self.first_bar == True:
            self.last_bar = self.time
            self.calc_timeframe = self.timeframe
            self.first_bar = False
            return True

        if self.current_bar != self.last_bar:
            self.last_bar = self.current_bar
            self.calc_timeframe = self.timeframe
            return True
        else:
            return False

    @property
    def is_new_tick(self) -> bool:
        """
        check if a new tick has occured
        """
        self.current_tick = pd.to_datetime(self.tick.time, unit='s')

        if self.first_tick == True:
            self.last_tick = pd.to_datetime(self.tick.time, unit='s')
            self.first_tick = False
            return True

        if self.current_tick != self.last_tick:
            self.last_tick = self.current_tick
            return True
        else:
            return False

    @property
    def yen_adjust_factor(self):
        """
        This function returns a constant factor, which is used for position sizing for Yen pairs
        """
        output = 1
        if self.digits == 2 or self.digits == 3:
            output = 100
        return output

    def get_ticks_from(self,
                       datetime_from: Union[datetime, int],
                       count: int,) -> pd.DataFrame:
        """
        get tick data as dataframe
        >>> datetime_from: starting date datetime
        >>> count: how many ticks you want
        """
        with self.connected:
            ticks = copy_ticks_from(
                self.name, datetime_from, count, const.COPY_TICKS.ALL)
            ticks_frame = pd.DataFrame(ticks)
            # convert time in seconds into the datetime format
            ticks_frame['time'] = pd.to_datetime(ticks_frame['time'], unit='s')
            return ticks_frame

    def get_rates_position(self):
        """
        This will get array of bar data, if you want dataframe use make_ohlcv function
        >>> `nth_bar`: how many bar data we want
        """
        with self.connected:
            if self.max_bars:
                # ! 1 for last candle, which is closed
                bars = copy_rates_from_pos(
                    self.name, self.timeframe, 1, self.max_bars)
            else:
                # ! 1 for last candle, which is closed
                bars = copy_rates_from_pos(
                    self.name, self.timeframe, 1, 99999)
            return bars

    def get_rates_bar(self, ohlcv: bool = True, datetime_from: Union[datetime, int] = None, datetime_to: Union[datetime, int] = None,):
        """
        This will get array of bar data, if you want dataframe use make_ohlcv function
        create date time like this 
        "date_string = "2022-01-15"
        date_format = "%Y-%m-%d"

        date_object = datetime.strptime(date_string, date_format)"
        If you want from to, don't use max bars when creating Symbol,
        IF YOU USE 'datetime_from', 'datetime_to' is also REQUIRED

        """
        with self.connected:
            if self.max_bars:
                # ! 1 for last candle, which is closed
                bars = copy_rates(self.name, self.timeframe,
                                  start_pos=1, count=self.max_bars, datetime_from=datetime_from, datetime_to=datetime_to)
            else:
                bars = copy_rates(
                    self.name, self.timeframe, datetime_from=datetime_from, datetime_to=datetime_to)

            if ohlcv:
                df = make_ohlcv(bars)
                return df
            else:
                return bars

    def normalize_price(self, price: float):
        """
        Normalizes prices as per digits of the pair
        """
        ts = self.trade_tick_size
        if ts != 0.0:
            return round(round(price / ts) * ts, self.digits)

    def lot_size(self, capital, risk_percent: float, stop_point: int, MAX_RISK_PERCENT: float, NEED_CONFIRMATION_FOR_BUY: bool, entry: Entry):
        """
        Find out lot size with stoploss and risk percentage
        """
        # assert stop_point != 0 or stop_point is not None, "You must define stop loss, whoever or what ever your strategy is, I won't allow it"

        # ! checking here if we have define stop_point and risk_percent
        if risk_percent > 0 and stop_point > 0:
            # ! if for some foolish reason we have defined the stop loss to be more than our max risk percent than our risk percent will be max
            trade_size = 0
            while trade_size == 0:
                if risk_percent > MAX_RISK_PERCENT:
                    risk_percent = MAX_RISK_PERCENT

                p = self.p_value

                # ! finding out account margin (MARGIN IS THE AMOUNT WILL TAKE AS LOSS WHEN OUR INITIAL STOP IS HIt)
                margin = capital * (risk_percent / 100)
                tick_size = self.trade_tick_value
                # ! stop_point * p to make a normal pip for both 4 digit and 5 digit broker
                # also with (risk_percent*0.01*account_balance)/(self.trade_contract_size *
                #                self.trade_tick_value*stop_point*self.point * p)
                if margin >= stop_point:
                    tick_size = tick_size * 100
                trade_size = (margin/(stop_point * p)) / tick_size
                # ! to take the lowest value, as round and floor makes it higher but to stay with my risk I will reduce it to lowest ex: 0.0356 to 0.03 not 0.04
                trade_size = float(str(trade_size)[:4])
                if trade_size == 0:
                    if not NEED_CONFIRMATION_FOR_BUY:
                        # ! if autotrade than we will not open a position
                        return 0, 0, 0, 0
                    # ! we will find a new stop loss
                    old_stop_point = stop_point
                    stop_point = (margin / tick_size) * p
                    print(
                        f"With the current stop point of {old_stop_point}, the risk percent of {risk_percent} will have {trade_size} trade size, the stop point must be max at {stop_point}, new stop point is changed at {int(stop_point)}")
                    try:
                        if entry == Entry.Buy:
                            new_price = self.add_subtract_pips_to_price(
                                self.bid, stop_point, add=False)
                            stop_point_agree = int(input(
                                f"do you agree to place stop loss at: {self.normalize_price(new_price)}, enter 1 for yes: "))

                            if stop_point_agree != 1:
                                stop_price = float(
                                    input("define the stop point where you want your stop loss: "))
                                stop_point = self.price_diff_to_pips(
                                    self.bid, stop_price, convert=True)

                                risk_percent = float(input(
                                    "Define a new increased risk percent, cause previous risk percent won't work for this trade: "))
                            else:
                                trade_size = self.volume_min
                        elif entry == Entry.Sell:
                            new_price = self.add_subtract_pips_to_price(
                                self.ask, stop_point, add=True)
                            stop_point_agree = int(input(
                                f"do you agree to place stop loss at: {self.normalize_price(new_price)}, enter 1 for yes: "))

                            if stop_point_agree != 1:
                                stop_price = float(
                                    input("define the stop point where you want your stop loss: "))
                                stop_point = self.price_diff_to_pips(
                                    self.ask, stop_price, convert=True)

                                risk_percent = float(input(
                                    "Define a new increased risk percent, cause previous risk percent won't work for this trade: "))
                            else:
                                trade_size = self.volume_min

                    except Exception as ve:
                        print(ve.args[0])
                        print(
                            "Typed wrong, try again")
                        stop_point = old_stop_point

            trade_size = self.verify_volume(trade_size)

        else:
            print("No stop point and risk percent is defined")
            trade_size = 0

        return entry, int(stop_point), trade_size, risk_percent

    def lot_size_bt(self, capital, risk_percent: float, stop_point: int, MAX_RISK_PERCENT: float):
        """
        Find out lot size with stoploss and risk percentage, this one is for backtesting purposes
        """
        # assert stop_point != 0 or stop_point is not None, "You must define stop loss, whoever or what ever your strategy is, I won't allow it"

        # ! checking here if we have define stop_point and risk_percent
        if risk_percent > 0 and stop_point > 0:
            # ! if for some foolish reason we have defined the stop loss to be more than our max risk percent than our risk percent will be max
            if risk_percent > MAX_RISK_PERCENT:
                risk_percent = MAX_RISK_PERCENT

            p = self.p_value

            # ! finding out account margin (MARGIN IS THE AMOUNT WILL TAKE AS LOSS WHEN OUR INITIAL STOP IS HIt)
            margin = capital * (risk_percent / 100)
            tick_size = self.trade_tick_value
            # ! stop_point * p to make a normal pip for both 4 digit and 5 digit broker
            # also with (risk_percent*0.01*account_balance)/(self.trade_contract_size *
            #                self.trade_tick_value*stop_point*self.point * p)
            trade_size = (margin/(stop_point * p)) / tick_size
            # ! to take the lowest value, as round and floor makes it higher but to stay with my risk I will reduce it to lowest ex: 0.0356 to 0.03 not 0.04
            trade_size = float(str(trade_size)[:4])
            if trade_size == 0:
                old_stop_point = stop_point
                stop_point = (margin / tick_size) * p
                print(
                    f"With the current stop point of {old_stop_point}, the risk percent of {risk_percent} will have {trade_size} trade size, the stop point must be max at {stop_point}, new stop point is changed at {int(stop_point)}")
            trade_size = self.verify_volume(trade_size)

        else:
            print("No stop point and risk percent is defined")
            trade_size = 0

        return int(stop_point), trade_size

    def verify_volume(self, lot) -> float:
        """
        verify volume with the lowest and highest volume,  size and step value and returns the newely calculated volume
        >>> lot: our calculated lot size
        """
        min_volume = self.volume_min
        max_volume = self.volume_max
        step_volume = self.volume_step

        if lot < min_volume:
            lot = min_volume
        elif lot > max_volume:
            lot = max_volume
        # ! making trade size with step volume
        else:
            lot = round(lot/step_volume) * step_volume

        if step_volume >= 0.1:
            lot = round(lot, 1)
        else:
            lot = round(lot, 2)

        return lot

    def add_subtract_pips_to_price(self, price, pips, add):
        """
        adds or subtracts pips to the price
        >>> `price`: the price with which we want to add or subt
        >>> `pips`: with how many pips we want to add or subtract
        >>> `add`: 'True' for add 'False' for subtract
        """
        P = self.p_value
        pips = (pips*self.point*P)
        if add:
            ret_price = price + pips
        else:
            ret_price = price - pips
        return ret_price

    def price_diff_to_pips(self, first_price: float, second_price: float, convert=False):
        """
        Calculate distance between two price points and convert the differece in pips
        symbol_info: symbol dictionary
        order_price: executed order price
        stop_price: our defined stop price
        convert: if true 20pip, otherwise 200pip(pip for 5 digit broker)
        """
        stop_diff = abs(first_price-second_price)
        get_point = self.point
        if convert:
            P = self.p_value
            price_to_point = int((stop_diff / get_point)/P)
            # ! exact pip like 20 pip
        else:
            price_to_point = int(stop_diff / get_point)
            # ! pip for 5 digit broker like 200

        return price_to_point

    def tick_calc(self, price: float, num_ticks: int):
        """Calculate a new price by number of ticks from the price param. The result is normalized to the
        tick-size of the instrument.

        :param price: The price to add or subtract ticks from.
        :param num_ticks: number of ticks. If subtracting ticks then this should be a negative number.
        :return: A new price adjusted by the number of ticks and normalized to tick-size.
        """
        return self.normalize_price(price + num_ticks * self.trade_tick_size)

    def refresh_rates(self):
        # with self.connected:
        self.tick
        return self

    def _refresh(self):
        info = self._info or self.symbol_info
        # ! if symbol is not selected in market watch select the symbol
        self.select = info.select if info.select else True
        self.refresh_rates()
        # self.spread = info.spread
        # self.volume_real = info.volume_real
        self.custom = info.custom
        self.chart_mode = info.chart_mode
        self.visible = info.visible
        self.session_deals = info.session_deals
        self.session_buy_orders = info.session_buy_orders
        self.session_sell_orders = info.session_sell_orders
        # self.volume = info.volume
        self.volumehigh = info.volumehigh
        self.volumelow = info.volumelow
        self.digits = info.digits
        self.spread_float = info.spread_float
        self.ticks_bookdepth = info.ticks_bookdepth
        self.trade_calc_mode = info.trade_calc_mode
        self.trade_mode = info.trade_mode
        self.start_time = info.start_time
        self.expiration_time = info.expiration_time
        self.trade_stops_level = info.trade_stops_level
        self.trade_freeze_level = info.trade_freeze_level
        self.trade_exemode = info.trade_exemode
        self.swap_mode = info.swap_mode
        self.swap_rollover3days = info.swap_rollover3days
        self.margin_hedged_use_leg = info.margin_hedged_use_leg
        self.expiration_mode = info.expiration_mode
        self.filling_mode = info.filling_mode
        self.order_mode = info.order_mode
        self.order_gtc_mode = info.order_gtc_mode
        self.option_mode = info.option_mode
        self.option_right = info.option_right
        self.bidhigh = info.bidhigh
        self.bidlow = info.bidlow
        self.askhigh = info.askhigh
        self.asklow = info.asklow
        self.lasthigh = info.lasthigh
        self.lastlow = info.lastlow
        self.volumehigh_real = info.volumehigh_real
        self.volumelow_real = info.volumelow_real
        self.option_strike = info.option_strike
        self.point = info.point
        self.trade_tick_value = info.trade_tick_value
        self.trade_tick_value_profit = info.trade_tick_value_profit
        self.trade_tick_value_loss = info.trade_tick_value_loss
        self.trade_tick_size = info.trade_tick_size
        self.trade_contract_size = info.trade_contract_size
        self.trade_accrued_interest = info.trade_accrued_interest
        self.trade_face_value = info.trade_face_value
        self.trade_liquidity_rate = info.trade_liquidity_rate
        self.volume_min = info.volume_min
        self.volume_max = info.volume_max
        self.volume_step = info.volume_step
        self.volume_limit = info.volume_limit
        self.swap_long = info.swap_long
        self.swap_short = info.swap_short
        self.margin_initial = info.margin_initial
        self.margin_maintenance = info.margin_maintenance
        self.session_volume = info.session_volume
        self.session_turnover = info.session_turnover
        self.session_interest = info.session_interest
        self.session_buy_orders_volume = info.session_buy_orders_volume
        self.session_sell_orders_volume = info.session_sell_orders_volume
        self.session_open = info.session_open
        self.session_close = info.session_close
        self.session_aw = info.session_aw
        self.session_price_settlement = info.session_price_settlement
        self.session_price_limit_min = info.session_price_limit_min
        self.session_price_limit_max = info.session_price_limit_max
        self.margin_hedged = info.margin_hedged
        self.price_change = info.price_change
        self.price_volatility = info.price_volatility
        self.price_theoretical = info.price_theoretical
        self.price_greeks_delta = info.price_greeks_delta
        self.price_greeks_theta = info.price_greeks_theta
        self.price_greeks_gamma = info.price_greeks_gamma
        self.price_greeks_vega = info.price_greeks_vega
        self.price_greeks_rho = info.price_greeks_rho
        self.price_greeks_omega = info.price_greeks_omega
        self.price_sensitivity = info.price_sensitivity
        self.basis = info.basis
        self.category = info.category
        self.currency_base = info.currency_base
        self.currency_profit = info.currency_profit
        self.currency_margin = info.currency_margin
        self.bank = info.bank
        self.description = info.description
        self.exchange = info.exchange
        self.formula = info.formula
        self.isin = info.isin
        self.page = info.page
        self.path = info.path
        return self


if __name__ == "__main__":

    # tasks = set()

    # class Strategy():
    #     """
    #     Strategy class example
    #     """

    #     def __init__(self, symbol: str):
    #         self.symbol = Symbol(symbol, TIMEFRAME_M1)

    #     async def next(self):
    #         """
    #         will perform next check
    #         """
    #         symbol = self.symbol
    #         # print("Called for symbol: ", symbol.name)

    #         # print("2nd Call for symbol: ", symbol.name)
    #         if symbol.is_new_tick:
    #             print(f"{symbol.name} New Tick: ",
    #                   pd.to_datetime(symbol.tick.time, unit='s'))
    #             print("Bid: ", symbol.normalize_price(symbol.tick.bid),
    #                   " Ask: ", symbol.normalize_price(symbol.tick.ask))

    #         new_bar = symbol.is_new_bar
    #         if new_bar:
    #             print(f"{symbol.name} New Bar: ", symbol.time)
    #         # print(tasks)
    #         task = asyncio.create_task(self.next(), name=symbol)
    #         tasks.add(task)

    # symbols = ['EURUSD', 'GBPUSD', 'AUDUSD', 'USDJPY', 'USDCHF']

    # # -----------------------------------------------------+
    # # * (START)  One way of doing it
    # # -----------------------------------------------------+

    # async def run_live():
    #     for symbol in symbols:
    #         strtg = Strategy(symbol)
    #         task = asyncio.create_task(strtg.next(), name=symbol)
    #         tasks.add(task)
    #     while True:
    #         done, _pending = await asyncio.wait(tasks, timeout=0.01)
    #         tasks.difference_update(done)
    #         print('--'*10)

    # async def async_main() -> None:
    #     try:
    #         await run_live()
    #     except asyncio.CancelledError:
    #         print('Process got cancelled')
    #         # ! if for some reason we cancel the progress, we will cancel all the remaining process
    #         for task in tasks:
    #             task.cancel()
    #         # ! for being grasious we will give one last chanche to any task might be pending
    #         done, pending = await asyncio.wait(tasks, timeout=1.0)
    #         # ! remvoing the done and any pending task in the list
    #         tasks.difference_update(done)
    #         tasks.difference_update(pending)
    #         # ! still checking if there is any pending task
    #         if tasks:
    #             print("warning: more tasks added while we were cancelling")

    # loop = asyncio.get_event_loop()
    # task = loop.create_task(async_main())
    # # ! after 10 seconds we will cancel the task
    # loop.call_later(600, task.cancel)
    # loop.run_until_complete(task)
    # -----------------------------------------------------+
    # * (END)  One way of doing it
    # -----------------------------------------------------+

    # sym = Symbol('EURJPY', const.TIMEFRAME_H1, 3000
    sym = Symbol('EURJPY', const.TIMEFRAME_H1)
    # ! WE CAN ALSO CHANGE THE TIMEFRAME, LAST BAR VALUE WILL BE CHANGED TOO
    # sym.timeframe = const.TIMEFRAME_M1
    # rates_df = sym.get_rates_bar()
    # df = make_ohlcv(rates_df)
    strt_date_string = "2023-01-01"
    end_date_string = "2023-02-13"
    date_format = "%Y-%m-%d"
    strt_dt = datetime.strptime(strt_date_string, date_format)
    end_dt = datetime.strptime(end_date_string, date_format)
    df = sym.get_rates_bar(datetime_from=strt_dt, datetime_to=end_dt)
    print(df)

    print(sym.high_bars)
    time = sym.time
    print(time)
    close = sym.normalize_price(sym.close)
    print(close)
    close_array = sym.close_bars.to_numpy()
    print(close_array[-2])
    low = sym.normalize_price(sym.low)
    print(low)
    sym.timeframe = TIMEFRAME_H4
    close_array = sym.close_bars.to_numpy()
    print(close_array[-2])
    new_bar = sym.is_new_bar
    tries = 0
    while tries < 6000:
        new_tick = sym.is_new_tick
        if new_tick:
            print("New Tick: ", pd.to_datetime(sym.tick.time, unit='s'))
            print("Bid: ", sym.normalize_price(sym.tick.bid),
                  " Ask: ", sym.normalize_price(sym.tick.ask))
        # else:
        #     print('Not a new tick, Old Tick: ',
        #           pd.to_datetime(sym.tick.time, unit='s'))

        new_bar = sym.is_new_bar
        if new_bar:
            print("New Bar: ", sym.time)
        # else:
        #     print("Old Bar: ", sym.time)

        tries += 1
    print(Symbol('CHFJPY', TIMEFRAME_H1))





from datetime import datetime
from typing import Union
from MT5TrdHelper.connect.connection import Connection
import pymt5adapter as mt5
from pymt5adapter.order import Order
import pandas as pd
from abc import abstractmethod, ABCMeta


class Exchaenge(Connection, metaclass=ABCMeta):
    def __init__(self,
                 path,
                 portable,
                 server,
                 login,
                 password,
                 timeout,
                 logger,  # default is None
                 ensure_trade_enabled,  # default is False
                 enable_real_trading,  # default is False
                 raise_on_errors,  # default is False
                 return_as_dict,  # default is False
                 return_as_native_python_objects   # default is False)
                 ):
        """
        exchange class for account re
        """
        super(Exchaenge, self).__init__(path,
                                        portable,
                                        server,
                                        login,
                                        password,
                                        timeout,
                                        logger,  # default is None
                                        ensure_trade_enabled,  # default is False
                                        enable_real_trading,  # default is False
                                        raise_on_errors,  # default is False
                                        return_as_dict,  # default is False
                                        return_as_native_python_objects   # default is False)
                                        )

    @abstractmethod
    def get_account_info(self):
        """
        Get current active account information from mt5
        """
        raise NotImplementedError

    @abstractmethod
    def get_position_from_order_send_result(self, result: mt5.OrderSendResult):
        """
        Function to get recently placed order details with ordersendresult object, which is returned by order_send function call
        """
        raise NotImplementedError

    @abstractmethod
    def get_position_with_ticket_id(self, ticket_id: int):
        """
        Get an open position with ticket id
        """
        raise NotImplementedError

    @abstractmethod
    def get_position_with_symbol(self, symbol: str):
        """
        Get an open position with ticket id
        """
        raise NotImplementedError

    @abstractmethod
    def get_position_with_magic(self, magic: int):
        """
        Get all open position with magic number
        """
        raise NotImplementedError

    @abstractmethod
    def get_all_open_position(self):
        """
        Get all open positions
        """
        raise NotImplementedError

    @abstractmethod
    def get_total_open_positions(self):
        """
        Get total number of open positions
        """
        raise NotImplementedError

    @abstractmethod
    def get_history_position(self, order_id: int):
        """
        Get a position from order history with order id
        """
        raise NotImplementedError

    @abstractmethod
    def market_order_send(self, symbol, volume, order_type, stop_loss, take_profit, slippage, magic, position=0):
        """
        Function to place a market order(open order, close order)
        """
        raise NotImplementedError

    @abstractmethod
    def modify_sl(self, stop_loss, position):
        """
        Modify stop loss
        """
        raise NotImplementedError

    @abstractmethod
    def modify_tp(self, take_profit, position):
        """
        Modify stop loss
        """
        raise NotImplementedError


if __name__ == "__main__":
    conn = mt5.connected()
    with conn:
        exc = Exchaenge()
        symbol_positions = exc.get_position_with_symbol('GBPUSD')
        print(symbol_positions)



import io
from datetime import datetime, timedelta
from localconfig.manager import LocalConfig
from typing import Union
from pymt5adapter import order
from pymt5adapter.const import COPY_TICKS, ORDER_TIME, ORDER_TYPE, POSITION_TYPE, TIMEFRAME_M1, TRADE_ACTION, ORDER_FILLING
from MT5TrdHelper.exchange.exchange_base import Exchaenge
import pymt5adapter as mt5
from pymt5adapter.order import Order
from pymt5adapter.types import TradePosition
from typing import Tuple
import pandas as pd

# connect_path = "D:\\DATA_SET_MACHINE_LEARNING\\trading_engine\\my_trading_system\\connect\\connection.ini"
# add_setting_path = "D:\\DATA_SET_MACHINE_LEARNING\\trading_engine\\my_trading_system\\connect\\additional_settings.ini"


class MT5Exchange(Exchaenge):
    def __init__(self, connect_path, add_setting_path, terminal='real_terminal_connection'):
        """
        exchange class for account re
        """
        # -----------------------------------------------------+
        # * (START)  reading all the configuration and initializing connection class
        # -----------------------------------------------------+
        coneection_config = LocalConfig()
        coneection_config.read(connect_path)
        # config.read('.\\connection.ini')
        keyargs = dict(coneection_config.items(terminal))

        add_sett_config = LocalConfig()
        add_sett_config.read(add_setting_path)
        self.ADDITIONAL_SETTINGS = dict(add_sett_config.items(terminal))

        super(MT5Exchange, self).__init__(**keyargs)
        # -----------------------------------------------------+
        # * (END)  reading all the configuration and initializing connection class
        # -----------------------------------------------------+

    def get_terminal_info(self):
        """
        Get current active account information from mt5
        """
        with self.connected:
            terminal_info = mt5.terminal_info()
            return terminal_info

    def get_account_info(self):
        """
        Get current active account information from mt5
        """
        with self.connected:
            account_info = mt5.account_info()
            return account_info

    def get_current_capital(self, use_balance: bool = False, use_equity: bool = False, use_free_margin: bool = False):
        """
        Get current equity based on parameter
        """
        acc_inf = self.get_account_info()
        if use_balance:
            return acc_inf.balance
        elif use_equity:
            return acc_inf.equity
        elif use_free_margin:
            return acc_inf.margin_free

    def get_position_from_order_send_result(self, result: mt5.OrderSendResult):
        """
        Function to get recently placed order details with ordersendresult object, which is returned by order_send function call
        """
        with self.connected:
            if result.order != 0:
                pos = self.get_position_with_ticket_id(result.order)
            # ! than we take the first item from the order, and take it's position_id(ticket id) and with it we
            elif result.request.position != 0:
                pos = self.get_position_with_ticket_id(result.request.position)
            return pos

    def get_position_with_ticket_id(self, ticket_id: int):
        """
        Get an open position with ticket id
        """
        with self.connected:
            # ! we take the first item from order, and take it's position_id(ticket id) and with it we
            pos = mt5.positions_get(ticket=ticket_id)
            if pos:
                return pos[0]
            else:
                return dict()

    def get_position_with_symbol(self, symbol: str):
        """
        Get an open position with ticket id
        """
        with self.connected:
            # ! we take the first item from order, and take it's position_id(ticket id) and with it we
            pos = mt5.positions_get(symbol=symbol)
            return pos

    def get_position_with_magic(self, magic: int):
        """
        Get all open position with magic number
        """
        positions = self.get_all_open_position()
        pos = filter(lambda p: p.magic == magic, positions)
        pos = list(pos)
        return pos

    def get_all_open_position(self) -> Tuple[TradePosition]:
        """
        Get all open positions
        >>> filter from all positions     
        pos = filter(lambda position: position.ticket == 206080473
                 and position.symbol == 'EURUSD', positions)
        pos = list(pos)
        """
        with self.connected:
            positions = mt5.positions_get()
            return positions

    def get_all_pending_position(self) -> Tuple[TradePosition]:
        """
        Get all open positions
        >>> filter from all positions     
        pos = filter(lambda position: position.ticket == 206080473
                 and position.symbol == 'EURUSD', positions)
        pos = list(pos)
        """
        with self.connected:
            positions = mt5.orders_get()
            positions = filter(lambda pos: pos.type !=
                               ORDER_TYPE.BUY or pos.type != ORDER_TYPE.SELL, positions)
            return tuple(positions)

    def get_total_open_positions(self, all_positions: bool = False, buy_positions: bool = False, sell_positions: bool = False, magic: int = 0, with_magic: bool = False) -> int:
        """
        Get total number of open positions
        >>> If with magic than total position, buy position, sell position will be calculated under this magic number
        >>> Without magic than total position, buy position, sell position will be caluclated for all open position
        """
        with self.connected:
            if with_magic:
                assert magic != 0, 'Please define magic number'
                open_positions = self.get_position_with_magic(magic)
                if all_positions:
                    return len(open_positions)
                elif buy_positions:
                    positions = filter(lambda p: p.type ==
                                       POSITION_TYPE.BUY, open_positions)
                    tot_buy = len(list(positions))
                    return tot_buy
                elif sell_positions:
                    positions = filter(lambda p: p.type ==
                                       POSITION_TYPE.SELL, open_positions)
                    tot_sell = len(list(positions))
                    return tot_sell
            else:
                open_positions = self.get_all_open_position()
                if all_positions:
                    return len(open_positions)
                elif buy_positions:
                    positions = filter(lambda p: p.type ==
                                       POSITION_TYPE.BUY, open_positions)
                    tot_buy = len(list(positions))
                    return tot_buy
                elif sell_positions:
                    positions = filter(lambda p: p.type ==
                                       POSITION_TYPE.SELL, open_positions)
                    tot_sell = len(list(positions))
                    return tot_sell

    def get_history_position(self, order_id: int):
        """
        Get a position from order history with order id, it will return tuple with two position one with entry data and another with exit and profit data
        """
        with self.connected:
            position = mt5.history_deals_get(position=order_id)
            return position

    async def market_order_send(self, symbol: str, price: Union[float, None], volume, order_type, slippage, magic, action=mt5.TRADE_ACTION.DEAL, stop_loss: float = None, take_profit: float = None,  position_ticket=0, comment=None, type_time: ORDER_TIME = None, expiration: datetime = None):
        """
        Function to place a market order(open order, close order)
        """
        with self.connected:
            result = mt5.order_send(action=action, type=order_type,
                                    symbol=symbol, price=price, volume=volume, sl=stop_loss, tp=take_profit, deviation=slippage, type_filling=ORDER_FILLING(
                                        int(self.ADDITIONAL_SETTINGS['execution_type'])), position=position_ticket, magic=magic, comment=comment, type_time=type_time, expiration=expiration)
            if not result.retcode == mt5.TRADE_RETCODE.DONE:
                print(
                    f"Some error occured, while executing order: {result.comment}")
            return result

    def modify_sl(self, stop_loss, position):
        """
        Modify stop loss
        >>> `stop_loss`: stop loss as price
        >>> `position`: position object
        """
        # assert stop_loss, 'Stop loss can\'t be empty'
        with self.connected:

            order = Order.as_modify_sltp(
                position, sl=stop_loss, tp=position.tp, comment=position.comment)
            pos = order.send()
            return pos

    def modify_tp(self, take_profit, position):
        """
        Modify stop loss
        """
        assert take_profit, 'Take profit can\'t be empty'
        with self.connected:

            order = Order.as_modify_sltp(
                position, sl=position.sl, tp=take_profit, comment=position.comment)
            pos = order.send()
            return pos

    def modify_sl_tp(self, take_profit, stop_loss, position):
        """
        Modify stop and take profit together
        >>> `take_profit`: take profit price
        >>> `stop_loss`: stop loss price
        """
        # assert take_profit, 'Take profit can\'t be empty'
        with self.connected:
            order = Order.as_modify_sltp(
                position, sl=stop_loss, tp=take_profit, comment=position.comment)
            pos = order.send()
            return pos

    def calc_order_margin(self, order_type: int, symbol: str, lot: float, price: float):
        """
        Calculate margin required to open a position
        """

        assert lot != 0, "Calculation of lot failed for some reason"
        with self.connected:
            margin = mt5.order_calc_margin(order_type, symbol, lot, price)
            return margin

    def calc_order_profit(self, order_type: int, symbol: str, lot: float, price_open: float, current_price):
        """
        Calculate margin required to open a position
        """
        assert lot != 0, "Calculation of lot failed for some reason"
        with self.connected:
            profit = mt5.order_calc_profit(
                order_type, symbol, lot, price_open, current_price)
            return profit

    def validate_order_margin(self, order_type: int, symbol: str, lot: float, price: float):
        req_margin = self.calc_order_margin(order_type, symbol, lot, price)
        free_margin = self.get_current_capital(use_free_margin=True)
        if free_margin > req_margin:
            return True
        else:
            return False


if __name__ == "__main__":
    exc = MT5Exchange(terminal='XM')
    pos = exc.get_position_with_ticket_id(215066002)
    # ! FOR testing remove async from function
    # result = exc.market_order_send("GBPUSD", action=TRADE_ACTION.PENDING, price=1.4004, volume=0.4, order_type=ORDER_TYPE.BUY_STOP,
    #                                slippage=10, magic=1003, stop_loss=0.0, take_profit=0.0, comment='HI', type_time=ORDER_TIME.GTC, expiration=None)

    # loop = io.get_event_loop()
    # tasks = exc.get_position_with_symbol('GBPUSD'), exc.get_all_open_position()
    # symbol_positions, ret = loop.run_until_complete(io.gather(*tasks))
    # print(symbol_positions)
    # print(ret)

    # ret = exc.get_all_open_position()
    # pos = filter(lambda position: position.ticket == 206080473
    #              and position.symbol == 'EURUSD', ret)

    # ticks = exc.get_ticks_from('EURUSD', datetime.today(), 10)
    # ticks.iloc[-1]



from setuptools import setup, find_packages

VERSION = '0.1.3'
DESCRIPTION = 'Python MT5 trade helpers'
LONG_DESCRIPTION = 'This package is a wrapper on Mt5 python module'

# Setting up
setup(
    # the name must match the folder name 'verysimplemodule'
    name="MT5TrdHelper",
    version=VERSION,
    author="Nazrul Islam",
    author_email="nazrulgithub@gmail.com",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    packages=find_packages(),
    # add any additional packages that
    install_requires=['pandas', 'python-dateutil', 'pymt5adapter',
                      'localconfig'],
    # needs to be installed along with your package. Eg: 'caer'

    keywords=['python', 'mt5 python', 'mt5 python algotrading',
              'forex algotrading', 'forex python'],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Education",
        "Programming Language :: Python :: 3",
        "Operating System :: Microsoft :: Windows",
    ]
)
